# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""
Base classes for developing a keras.Model() Federated Learning model.

You may copy this file as the starting point of your own keras model.
"""

import copy
from warnings import catch_warnings, simplefilter

import numpy as np

from openfl.federated.task.runner import TaskRunner
from openfl.utilities import Metric, TensorKey, change_tags
from openfl.utilities.split import split_tensor_dict_for_holdouts

with catch_warnings():
    simplefilter(action="ignore")
    import keras

import logging

logger = logging.getLogger(__name__)


class KerasTaskRunner(TaskRunner):
    """The base model for Keras models in the federation.

    Attributes:
        model (keras.Model): The Keras model.
        model_tensor_names (list): List of model tensor names.
        required_tensorkeys_for_function (dict): A map of all of the required
            tensors for each of the public functions in KerasTaskRunner.
    """

    def __init__(self, **kwargs):
        """Initializes the KerasTaskRunner instance.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        self.model = keras.models.Model()

        self.model_tensor_names = []

        # this is a map of all of the required tensors for each of the public
        # functions in KerasTaskRunner
        self.required_tensorkeys_for_function = {}

    def rebuild_model(self, round_num, input_tensor_dict, validation=False):
        """Parse tensor names and update weights of model. Handles the
        optimizer treatment.

        Args:
            round_num (int): The round number.
            input_tensor_dict (dict): The input tensor dictionary.
            validation (bool, optional): If True, validate the model. Defaults
                to False.
        """
        if self.opt_treatment == "RESET":
            self.reset_opt_vars()
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)
        elif round_num > 0 and self.opt_treatment == "CONTINUE_GLOBAL" and not validation:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=True)
        else:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)

    def train_task(
        self,
        col_name,
        round_num,
        input_tensor_dict,
        metrics,
        epochs=1,
        batch_size=1,
        **kwargs,
    ):
        """
        Perform the training. Is expected to perform draws randomly, without
        replacement until data is exhausted. Then data is replaced and shuffled
        and draws continue.

        Args:
            col_name (str): The collaborator name.
            round_num (int): The round number.
            input_tensor_dict (dict): The input tensor dictionary.
            metrics (list): List of metrics.
            epochs (int, optional): Number of epochs to train. Defaults to 1.
            batch_size (int, optional): Batch size. Defaults to 1.
            **kwargs: Additional parameters.

        Returns:
            global_tensor_dict (dict): Dictionary of 'TensorKey: nparray'.
            local_tensor_dict (dict): Dictionary of 'TensorKey: nparray'.
        """
        if metrics is None:
            raise KeyError("metrics must be defined")

        # rebuild model with updated weights
        self.rebuild_model(round_num, input_tensor_dict)
        for epoch in range(epochs):
            logger.info("Run %s epoch of %s round", epoch, round_num)
            results = self.train_(
                self.data_loader.get_train_loader(batch_size),
                metrics=metrics,
                **kwargs,
            )

        # output metric tensors (scalar)
        origin = col_name
        tags = ("trained",)
        output_metric_dict = {
            TensorKey(metric_name, origin, round_num, True, ("metric",)): metric_value
            for (metric_name, metric_value) in results
        }

        # output model tensors (Doesn't include TensorKey)
        output_model_dict = self.get_tensor_dict(
            with_opt_vars=(self.opt_treatment == "CONTINUE_GLOBAL")
        )
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )

        # create global tensorkeys
        global_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags): nparray
            for tensor_name, nparray in global_model_dict.items()
        }
        # create tensorkeys that should stay local
        local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags): nparray
            for tensor_name, nparray in local_model_dict.items()
        }
        # the train/validate aggregated function of the next round will look
        # for the updated model parameters.
        # this ensures they will be resolved locally
        next_local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num + 1, False, ("model",)): nparray
            for tensor_name, nparray in local_model_dict.items()
        }

        global_tensor_dict = {
            **output_metric_dict,
            **global_tensorkey_model_dict,
        }
        local_tensor_dict = {
            **local_tensorkey_model_dict,
            **next_local_tensorkey_model_dict,
        }

        # update the required tensors if they need to be pulled from the
        # aggregator
        # TODO this logic can break if different collaborators have different
        # roles between rounds.
        # for example, if a collaborator only performs validation in the first
        # round but training in the second, it has no way of knowing the
        # optimizer state tensor names to request from the aggregator because
        # these are only created after training occurs. A work around could
        # involve doing a single epoch of training on random data to get the
        # optimizer names, and then throwing away the model.
        if self.opt_treatment == "CONTINUE_GLOBAL":
            self.initialize_tensorkeys_for_functions(with_opt_vars=True)

        return global_tensor_dict, local_tensor_dict

    def _initialize_metrics_result(self, batch_size):
        # evaluation needed before metrics can be resolved
        self.model.evaluate(self.data_loader.get_valid_loader(batch_size), verbose=1)
        return self.model.get_metrics_result()

    def train_(self, batch_generator, metrics: list = None, **kwargs):
        """Train single epoch. Override this function for custom training.

        Args:
            batch_generator (generator): Generator of training batches.
            metrics (list, optional): Names of metrics to save. Defaults to
                None.
            **kwargs: Additional parameters.

        Returns:
            results (list): List of Metric objects.
        """
        if metrics is None:
            metrics = []
        # TODO Currently assuming that all metrics are defined at
        #  initialization (build_model).
        #  If metrics are added (i.e. not a subset of what was originally
        #  defined) then the model must be recompiled.

        batch_size = kwargs.get("batch_size", 1)
        try:
            results = self.model.get_metrics_result()
            if len(results) == 0:
                results = self._initialize_metrics_result(batch_size)
        except ValueError:
            self._initialize_metrics_result(batch_size)
            results = self.model.get_metrics_result()

        # TODO if there are new metrics in the flplan that were not included
        #  in the originally
        #  compiled model, that behavior is not currently handled.
        for param in metrics:
            if param not in results:
                raise ValueError(
                    f"KerasTaskRunner does not support specifying new metrics. "
                    f"Param_metrics = {metrics}"
                )

        history = self.model.fit(batch_generator, verbose=2, **kwargs)
        results = []
        for metric in metrics:
            value = np.mean([history.history[metric]])
            results.append(Metric(name=metric, value=np.array(value)))
        return results

    def validate_task(self, col_name, round_num, input_tensor_dict, **kwargs):
        """Run the trained model on validation data; report results.

        Args:
            col_name (str): The collaborator name.
            round_num (int): The round number.
            input_tensor_dict (dict): The input tensor dictionary. Either the
                last aggregated or locally trained model
            **kwargs: Additional parameters.

        Returns:
            output_tensor_dict (dict): Dictionary of 'TensorKey: nparray'.
                These correspond to acc, precision, f1_score, etc.
            dict: Empty dictionary.
        """
        batch_size = kwargs.get("batch_size", 1)

        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        param_metrics = kwargs["metrics"]

        self.model.evaluate(self.data_loader.get_valid_loader(batch_size), verbose=1)
        results = self.model.get_metrics_result()

        # TODO if there are new metrics in the flplan that were not included in
        #  the originally compiled model, that behavior is not currently
        #  handled.
        for param in param_metrics:
            if param not in results:
                raise ValueError(
                    f"KerasTaskRunner does not support specifying new metrics. "
                    f"Param_metrics = {param_metrics}"
                )

        origin = col_name
        suffix = "validate"
        if kwargs["apply"] == "local":
            suffix += "_local"
        else:
            suffix += "_agg"
        tags = ("metric",)
        tags = change_tags(tags, add_field=suffix)
        output_tensor_dict = {
            TensorKey(metric, origin, round_num, True, tags): np.array(results[metric])
            for metric in param_metrics
        }

        return output_tensor_dict, {}

    def save_native(self, filepath):
        """Save model.

        Args:
            filepath (str): The file path to save the model.
        """
        self.model.export(filepath)

    def load_native(self, filepath):
        """Load model.

        Args:
            filepath (str): The file path to load the model.
        """
        self.model = keras.models.load_model(filepath)

    @staticmethod
    def _get_weights_names(obj):
        """Get the list of weight names.

        Args:
            obj (Model or Optimizer): The target object that we want to get
                the weights.

        Returns:
            weight_names (list): The weight name list.
        """
        if isinstance(obj, keras.optimizers.Optimizer):
            weight_names = [weight.name for weight in obj.variables]
        else:
            weight_names = [
                layer.name + "/" + weight.name for layer in obj.layers for weight in layer.weights
            ]
        return weight_names

    @staticmethod
    def _get_weights_dict(obj, suffix=""):
        """
        Get the dictionary of weights.

        Args:
            obj (Model or Optimizer): The target object that we want to get
                the weights.
            suffix (str, optional): Suffix for weight names. Defaults to ''.

        Returns:
            weights_dict (dict): The weight dictionary.
        """
        weights_dict = {}
        weight_names = KerasTaskRunner._get_weights_names(obj)
        if isinstance(obj, keras.optimizers.Optimizer):
            weights_dict = {
                weight_names[i] + suffix: weight.numpy()
                for i, weight in enumerate(copy.deepcopy(obj.variables))
            }
        else:
            weight_name_index = 0
            for layer in obj.layers:
                if weight_name_index < len(weight_names) and len(layer.get_weights()) > 0:
                    for weight in layer.get_weights():
                        weights_dict[weight_names[weight_name_index] + suffix] = weight
                        weight_name_index += 1
        return weights_dict

    @staticmethod
    def _set_weights_dict(obj, weights_dict):
        """Set the object weights with a dictionary.

        Args:
            obj (Model or Optimizer): The target object that we want to set
                the weights.
            weights_dict (dict): The weight dictionary.
        """
        weight_names = KerasTaskRunner._get_weights_names(obj)
        weight_values = [weights_dict[name] for name in weight_names]
        obj.set_weights(weight_values)

    def get_tensor_dict(self, with_opt_vars, suffix=""):
        """
        Get the model weights as a tensor dictionary.

        Args:
            with_opt_vars (bool): If we should include the optimizer's status.
            suffix (str): Universally.

        Returns:
            model_weights (dict): The tensor dictionary.
        """
        model_weights = self._get_weights_dict(self.model, suffix)

        if with_opt_vars:
            opt_weights = self._get_weights_dict(self.model.optimizer, suffix)

            model_weights.update(opt_weights)
            if len(opt_weights) == 0:
                logger.debug("WARNING: We didn't find variables for the optimizer.")
        return model_weights

    def set_tensor_dict(self, tensor_dict, with_opt_vars):
        """Set the model weights with a tensor dictionary.

        Args:
            tensor_dict (dict): The tensor dictionary.
            with_opt_vars (bool): True = include the optimizer's status.
        """
        if with_opt_vars is False:
            # It is possible to pass in opt variables from the input tensor
            # dict. This will make sure that the correct layers are updated
            model_weight_names = self._get_weights_names(self.model)
            model_weights_dict = {name: tensor_dict[name] for name in model_weight_names}
            self._set_weights_dict(self.model, model_weights_dict)
        else:
            model_weight_names = self._get_weights_names(self.model)
            model_weights_dict = {name: tensor_dict[name] for name in model_weight_names}
            opt_weight_names = self._get_weights_names(self.model.optimizer)
            opt_weights_dict = {name: tensor_dict[name] for name in opt_weight_names}
            self._set_weights_dict(self.model, model_weights_dict)
            self._set_weights_dict(self.model.optimizer, opt_weights_dict)

    def reset_opt_vars(self):
        """Resets the optimizer variables."""
        pass

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """Get the required tensors for specified function that could be called
        as part of a task.

        By default, this is just all of the layers and optimizer of the model.

        Args:
            func_name (str): The function name.
            **kwargs: Any function arguments.

        Returns:
            list: List of TensorKey objects.
        """
        if func_name == "validate_task":
            local_model = "apply=" + str(kwargs["apply"])
            return self.required_tensorkeys_for_function[func_name][local_model]
        else:
            return self.required_tensorkeys_for_function[func_name]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """Set the required tensors for all publicly accessible methods that
        could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function

        Args:
            with_opt_vars (bool, optional): If True, include the optimizer's
                status. Defaults to False.
        """
        # TODO there should be a way to programmatically iterate through all
        #  of the methods in the class and declare the tensors.
        # For now this is done manually

        output_model_dict = self.get_tensor_dict(with_opt_vars=with_opt_vars)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )
        if not with_opt_vars:
            global_model_dict_val = global_model_dict
            local_model_dict_val = local_model_dict
        else:
            output_model_dict = self.get_tensor_dict(with_opt_vars=False)
            global_model_dict_val, local_model_dict_val = split_tensor_dict_for_holdouts(
                output_model_dict,
                **self.tensor_dict_split_fn_kwargs,
            )

        self.required_tensorkeys_for_function["train_task"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function["train_task"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict
        ]

        # Validation may be performed on local or aggregated (global) model,
        # so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function["validate_task"] = {}
        # TODO This is not stateless. The optimizer will not be
        self.required_tensorkeys_for_function["validate_task"]["apply=local"] = [
            TensorKey(tensor_name, "LOCAL", 0, False, ("trained",))
            for tensor_name in {**global_model_dict_val, **local_model_dict_val}
        ]
        self.required_tensorkeys_for_function["validate_task"]["apply=global"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict_val
        ]
        self.required_tensorkeys_for_function["validate_task"]["apply=global"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict_val
        ]



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Utilities module."""

import hashlib
import ipaddress
import logging
import os
import re
import shutil
import stat
from collections.abc import Callable
from functools import partial
from socket import getfqdn
from typing import List, Optional, Tuple

from dynaconf import Dynaconf
from tqdm import tqdm


def getfqdn_env(name: str = "") -> str:
    """
    Get the system FQDN, with priority given to environment variables.

    This function retrieves the fully qualified domain name (FQDN) of the
    system.
    If the 'FQDN' environment variable is set, its value is returned.
    Otherwise,the FQDN is determined based on the system's hostname.

    Args:
        name (str, optional): The name from which to extract the FQDN.
            Defaults to ''.

    Returns:
        str: The FQDN of the system.
    """
    fqdn = os.environ.get("FQDN", None)
    if fqdn is not None:
        return fqdn
    return getfqdn(name)


def is_fqdn(hostname: str) -> bool:
    """Check if a hostname is a fully qualified domain name.

    This function checks if a hostname is a fully qualified domain name (FQDN)
    according to the rules specified on Wikipedia.
    https://en.m.wikipedia.org/wiki/Fully_qualified_domain_name.

    Args:
        hostname (str): The hostname to check.

    Returns:
        bool: `True` if the hostname is a FQDN, `False` otherwise.
    """
    if not 1 < len(hostname) < 253:
        return False

    # Remove trailing dot
    hostname.rstrip(".")

    #  Split hostname into list of DNS labels
    labels = hostname.split(".")

    #  Define pattern of DNS label
    #  Can begin and end with a number or letter only
    #  Can contain hyphens, a-z, A-Z, 0-9
    #  1 - 63 chars allowed
    fqdn = re.compile(r"^[a-z0-9]([a-z-0-9-]{0,61}[a-z0-9])?$", re.IGNORECASE)  # noqa FS003

    # Check that all labels match that pattern.
    return all(fqdn.match(label) for label in labels)


def is_api_adress(address: str) -> bool:
    """Validate IP address value.

    This function checks if a string is a valid IP address.

    Args:
        address (str): The string to check.

    Returns:
        bool: `True` if the string is a valid IP address, `False` otherwise.
    """
    try:
        ipaddress.ip_address(address)
        return True
    except ValueError:
        return False


def add_log_level(level_name, level_num, method_name=None):
    """Add a new logging level to the logging module.

    This function adds a new logging level to the logging module with a
    specified name, value, and method name.

    Args:
        level_name (str): The name of the new logging level.
        level_num (int): The value of the new logging level.
        method_name (str, optional): The name of the method to use for
            the new logging level. Defaults to None.
    """
    if not method_name:
        method_name = level_name.lower()

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def validate_file_hash(file_path, expected_hash, chunk_size=8192):
    """Validate SHA384 hash for file specified.

    This function validates the SHA384 hash of a file against an expected hash.

    Args:
        file_path (str): The path to the file to validate.
            (absolute or relative to the current working directory) of the file
            to be opened or an integer file descriptor of the file to be
            wrapped.
        expected_hash (str): The expected SHA384 hash of the file.
        chunk_size (int, optional): The size of the chunks to read from the
            file. Defaults to 8192.

    Raises:
        SystemError: If the hash of the file does not match the expected hash.
    """
    h = hashlib.sha384()
    with open(file_path, "rb") as file:
        # Reading is buffered, so we can read smaller chunks.
        while True:
            chunk = file.read(chunk_size)
            if not chunk:
                break
            h.update(chunk)

    if h.hexdigest() != expected_hash:
        raise SystemError("ZIP File hash doesn't match expected file hash.")


def tqdm_report_hook():
    """Visualize downloading.

    This function creates a progress bar for visualizing the progress of a
    download.

    Returns:
        Callable: A function that updates the progress bar.
    """

    def report_hook(pbar, count, block_size, total_size):
        """Update progressbar."""
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = count * block_size
        pbar.update(progress_bytes - pbar.n)

    pbar = tqdm(total=None)
    return partial(report_hook, pbar)


def merge_configs(
    overwrite_dict: Optional[dict] = None,
    value_transform: Optional[List[Tuple[str, Callable]]] = None,
    **kwargs,
) -> Dynaconf:
    """
    Create Dynaconf settings, merge its with `overwrite_dict` and validate
    result.

    This function creates a Dynaconf settings object, merges it with an
    optional dictionary, applies an optional value transformation, and
    validates the result.

    Args:
        overwrite_dict (Optional[dict], optional): A dictionary to merge with
            the settings. Defaults to None.
        value_transform (Optional[List[Tuple[str, Callable]]], optional): A
            list of tuples, each containing a key and a function to apply to
            the value of that key. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the Dynaconf
            constructor.

    Returns:
        Dynaconf: The merged and validated settings.
    """
    settings = Dynaconf(**kwargs, YAML_LOADER="safe_load")
    if overwrite_dict:
        for key, value in overwrite_dict.items():
            if value is not None or settings.get(key) is None:
                settings.set(key, value, merge=True)
    if value_transform:
        for key, operation in value_transform:
            value = settings.get(key)
            settings.set(key, operation(value))
    settings.validators.validate()
    return settings


def change_tags(tags, *, add_field=None, remove_field=None) -> Tuple[str, ...]:
    """Change tensor tags to add or remove fields.

    This function adds or removes fields from tensor tags.

    Args:
        tags (Tuple[str, ...]): The tensor tags.
        add_field (str, optional): A new tensor tag field to add. Defaults to
            None.
        remove_field (str, optional): A tensor tag field to remove. Defaults
            to None.

    Returns:
        Tuple[str, ...]: The modified tensor tags.

    Raises:
        Exception: If `remove_field` is not in `tags`.
    """
    tags = list(set(tags))

    if add_field is not None and add_field not in tags:
        tags.append(add_field)
    if remove_field is not None:
        if remove_field in tags:
            tags.remove(remove_field)
        else:
            raise Exception(f"{remove_field} not in tags {tuple(tags)}")

    tags = tuple(sorted(tags))
    return tags


def rmtree(path, ignore_errors=False):
    """Remove a directory tree.

    This function removes a directory tree. If a file in the directory tree is
    read-only, its read-only attribute is cleared before it is removed.

    Args:
        path (str): The path to the directory tree to remove.
        ignore_errors (bool, optional): Whether to ignore errors. Defaults to
            False.

    Returns:
        str: The path to the removed directory tree.
    """

    def remove_readonly(func, path, _):
        "Clear the readonly bit and reattempt the removal"
        if os.name == "nt":
            os.chmod(path, stat.S_IWRITE)  # Windows can not remove read-only files.
        func(path)

    return shutil.rmtree(path, ignore_errors=ignore_errors, onerror=remove_readonly)


def generate_port(hash, port_range=(49152, 60999)):
    """
    Generate a deterministic port number based on a hash and a unique key.

    Args:
        hash (str): A string representing the hash of the plan.
        port_range (tuple): A tuple containing the minimum and maximum port
            numbers (inclusive). The default range is (49152, 60999).

    Returns:
        int: A port number within the specified range.
    """
    min_port, max_port = port_range
    # Use the first 8 characters of the unique hash to ensure deterministic output
    hash_segment = hash[:8]
    port = int(hash_segment, 16) % (max_port - min_port) + min_port
    return port



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import os
import signal
import subprocess
import sys
from logging import getLogger

import psutil

from openfl.transport.grpc.interop import FlowerInteropClient
from openfl.utilities.path_check import is_directory_traversal

logger = getLogger(__name__)


class ConnectorFlower:
    """
    A Connector subclass specifically designed for integrating with the Flower framework.
    This class is responsible for constructing and managing the execution of Flower server commands.
    """

    def __init__(
        self,
        superlink_host,
        fleet_api_port,
        exec_api_port,
        serverappio_api_port,
        insecure=True,
        flwr_app_name=None,
        federation_name=None,
        automatic_shutdown=True,
        flwr_dir=None,
        **kwargs,
    ):
        """
        Initialize the ConnectorFlower instance by setting up the necessary server commands.

        Args:
            superlink_host (str): Host address for the Flower SuperLink.
            fleet_api_port (int): Port for the fleet API.
            exec_api_port (int): Port for the exec API.
            serverappio_api_port (int): Port for the serverappio API.
            insecure (bool): Whether to use insecure connections. Defaults to True.
            flwr_app_name (str, optional): Name of the Flower application to run. Defaults to None.
            federation_name (str, optional): Name of the federation. Defaults to None.
            automatic_shutdown (bool, optional): Whether to enable automatic shutdown.
                Defaults to True.
            flwr_dir (str, optional): Directory for Flower app within the OpenFL workspace.
                Plan.yaml configuration defaults to `save/.flwr`
            **kwargs: Additional keyword arguments.
        """
        super().__init__()
        self._process = None

        self.flwr_dir = flwr_dir
        if is_directory_traversal(self.flwr_dir):
            logger.error("Flower app directory path is out of the OpenFL workspace scope.")
            sys.exit(1)
        else:
            os.makedirs(self.flwr_dir, exist_ok=True)
            os.environ["FLWR_HOME"] = self.flwr_dir

        self.automatic_shutdown = automatic_shutdown
        self.signal_shutdown_sent = False

        self.superlink_params = {
            "insecure": insecure,
            "exec_api_port": exec_api_port,
            "fleet_api_port": fleet_api_port,
            "serverappio_api_port": serverappio_api_port,
        }
        self.superlink_host = superlink_host
        self.flwr_superlink_command = self._build_flwr_superlink_command()

        if flwr_app_name is None or federation_name is None:
            self.flwr_run_params = None
        else:
            self.flwr_run_params = {
                "flwr_app_name": flwr_app_name,
                "federation_name": federation_name,
            }
        self.flwr_run_command = self._build_flwr_run_command() if self.flwr_run_params else None

        self.interop_client = None
        signal.signal(signal.SIGINT, self._handle_sigint)

    def get_interop_client(self):
        """
        Create and return a FlowerInteropClient instance using the superlink parameters.

        Returns:
            FlowerInteropClient: An instance configured with the connector address
            and server rounds.
        """
        connector_port = self.superlink_params.get("fleet_api_port")
        connector_address = f"{self.superlink_host}:{connector_port}"
        self.interop_client = FlowerInteropClient(connector_address, self.automatic_shutdown)
        return self.interop_client

    def _build_flwr_superlink_command(self) -> list[str]:
        """
        Construct the command to initiate the Flower SuperLink based on provided parameters.

        Returns:
            list[str]: A list of command-line arguments for starting the Flower server.
        """

        command = ["flower-superlink", "--fleet-api-type", "grpc-adapter"]

        if self.superlink_params.get("insecure"):
            command += ["--insecure"]

        serverappio_api_port = self.superlink_params.get("serverappio_api_port")
        serverappio_api_address = f"{self.superlink_host}:{serverappio_api_port}"
        command += ["--serverappio-api-address", serverappio_api_address]

        fleet_api_port = self.superlink_params.get("fleet_api_port")
        fleet_api_address = f"{self.superlink_host}:{fleet_api_port}"
        command += ["--fleet-api-address", fleet_api_address]

        exec_api_port = self.superlink_params.get("exec_api_port")
        exec_api_address = f"{self.superlink_host}:{exec_api_port}"
        command += ["--exec-api-address", exec_api_address]

        if self.automatic_shutdown:
            command += ["--isolation", "process"]
            self.flwr_serverapp_command = self._build_flwr_serverapp_command()
            # flwr will default to "--isolation subprocess"

        return command

    def _build_flwr_serverapp_command(self) -> list[str]:
        """
        Construct the command to start the Flower ServerApp based on superlink parameters.

        Returns:
            list[str]: A list of command-line arguments for starting the Flower ServerApp.
        """
        command = ["flwr-serverapp", "--run-once"]

        if self.superlink_params["insecure"]:
            command += ["--insecure"]

        serverappio_api_port = self.superlink_params["serverappio_api_port"]
        serverappio_api_address = f"{self.superlink_host}:{serverappio_api_port}"
        command += ["--serverappio-api-address", serverappio_api_address]

        return command

    def is_flwr_serverapp_running(self):
        """
        Determine if the Flower ServerApp subprocess is currently active.

        Returns:
            bool: True if the ServerApp is running, False otherwise.
        """
        if not hasattr(self, "flwr_serverapp_subprocess"):
            logger.debug("[OpenFL Connector] ServerApp was never started.")
            return False

        if self.flwr_serverapp_subprocess.poll() is None:
            logger.debug("[OpenFL Connector] ServerApp is still running.")
            return True

        if not self.signal_shutdown_sent:
            self.signal_shutdown_sent = True
            logger.info(
                "[OpenFL Connector] Experiment has ended. Sending signal "
                "to shut down Flower components."
            )

        return False

    def _stop_flwr_serverapp(self):
        """Terminate the `flwr_serverapp` subprocess if it is still active."""
        if (
            hasattr(self, "flwr_serverapp_subprocess")
            and self.flwr_serverapp_subprocess.poll() is None
        ):
            logger.debug("[OpenFL Connector] ServerApp still running. Stopping...")
            self.flwr_serverapp_subprocess.terminate()
            try:
                self.flwr_serverapp_subprocess.wait(timeout=5)
            except subprocess.TimeoutExpired:
                self.flwr_serverapp_subprocess.kill()

    def _build_flwr_run_command(self) -> list[str]:
        """
        Construct the `flwr run` command to execute the Flower application.

        Returns:
            list[str]: A list of command-line arguments for running the Flower application.
        """
        federation_name = self.flwr_run_params.get("federation_name")
        flwr_app_name = self.flwr_run_params.get("flwr_app_name")

        command = ["flwr", "run", f"./src/{flwr_app_name}"]

        if federation_name:
            command.append(federation_name)

        return command

    def start(self):
        """
        Launch the `flower-superlink` and `flwr run` subprocesses
        using the constructed commands.
        """
        if self._process is None:
            logger.info(
                f"[OpenFL Connector] Starting server process: "
                f"{' '.join(self.flwr_superlink_command)}"
            )
            self._process = subprocess.Popen(self.flwr_superlink_command)
            logger.info(f"[OpenFL Connector] Server process started with PID: {self._process.pid}")
        else:
            logger.info("[OpenFL Connector] Server process is already running.")

        if hasattr(self, "flwr_run_command") and self.flwr_run_command:
            logger.info(
                f"[OpenFL Connector] Starting `flwr run` "
                f"subprocess: {' '.join(self.flwr_run_command)}"
            )
            subprocess.run(self.flwr_run_command)

        if hasattr(self, "flwr_serverapp_command") and self.flwr_serverapp_command:
            logger.info(
                f"[OpenFL Connector] Starting server app subprocess: "
                f"{' '.join(self.flwr_serverapp_command)}"
            )
            self.interop_client.set_is_flwr_serverapp_running_callback(
                self.is_flwr_serverapp_running
            )
            self.flwr_serverapp_subprocess = subprocess.Popen(self.flwr_serverapp_command)

    def stop(self):
        """Terminate the `flower-superlink` subprocess and any associated processes."""
        self._stop_flwr_serverapp()
        if self._process:
            try:
                logger.info(
                    f"[OpenFL Connector] Stopping server process with PID: {self._process.pid}..."
                )
                main_process = psutil.Process(self._process.pid)
                sub_processes = main_process.children(recursive=True)
                for sub_process in sub_processes:
                    logger.info(
                        (
                            f"[OpenFL Connector] Stopping server subprocess "
                            f"with PID: {sub_process.pid}..."
                        )
                    )
                    sub_process.terminate()
                _, still_alive = psutil.wait_procs(sub_processes, timeout=1)
                for p in still_alive:
                    p.kill()
                try:
                    self._process.terminate()
                    self._process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    self._process.kill()
                self._process = None
                logger.info("[OpenFL Connector] Server process stopped.")
            except Exception as e:
                logger.debug(f"[OpenFL Connector] Error during graceful shutdown: {e}")
                self._process.kill()
                logger.info("[OpenFL Connector] Server process forcefully terminated.")
        else:
            logger.info("[OpenFL Connector] No server process is currently running.")

    def _handle_sigint(self, signum, frame):
        """Handle the SIGINT signal (Ctrl+C) to cleanly stop the server process and its children."""
        logger.info("[OpenFL Connector] SIGINT received. Terminating server process...")
        self.stop()
        sys.exit(0)



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import hashlib
import logging
from importlib import util
from typing import Callable

from openfl.federated.data.sources.data_source import DataSource, DataSourceType

logging.getLogger("azure").setLevel(logging.WARNING)
logging.getLogger("azure.storage").setLevel(logging.WARNING)


class AzureBlobDataSource(DataSource):
    """Class for Azure Blob data"""

    def __init__(
        self,
        name: str,
        connection_string: str,
        container_name: str,
        folder_prefix="",
        hash_func: Callable[..., "hashlib._Hash"] = hashlib.sha384,
    ):
        # Check if azure.storage.blob is installed.
        if util.find_spec("azure.storage.blob") is None:
            raise Exception(
                "'azure-storage-blob' not installed."
                "This package is necessary for interacting with Azure Blob Storage."
            )
        from azure.storage.blob import BlobServiceClient

        super().__init__(DataSourceType.AZURE_BLOB, name)
        self.connection_string = connection_string
        self.container_name = container_name
        self.folder_prefix = folder_prefix
        if not super().is_valid_hash_function(hash_func):
            raise ValueError(
                f"Data source {self.name}: Invalid hash function: {hash_func.__name__}."
                " Must be a hashlib function."
            )
        self.hash_func = hash_func
        self._service_client = BlobServiceClient.from_connection_string(connection_string)
        self._container_client = self._service_client.get_container_client(container_name)

    def enumerate_files(self):
        """List all blobs in the container (full recursive listing)"""
        return [
            blob.name
            for blob in self._container_client.list_blobs(name_starts_with=self.folder_prefix)
        ]

    def compute_file_hash(self, blob_path: str):
        """Compute the hash of a blob's binary content."""
        content = self.read_blob(blob_path)
        return self.hash_func(content).hexdigest()

    def read_blob(self, blob_path: str):
        """Read blob content as binary data."""
        blob_client = self._container_client.get_blob_client(blob_path)
        downloader = blob_client.download_blob()
        return downloader.readall()

    @classmethod
    def from_dict(cls, ds_dict: dict):
        hash_func = getattr(hashlib, ds_dict.get("hash_func", "sha384"), None)
        return cls(
            name=ds_dict["name"],
            connection_string=ds_dict["connection_string"],
            container_name=ds_dict["container_name"],
            hash_func=hash_func,
        )



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Custom Assigner module."""

import logging
from collections import defaultdict

from openfl.interface.aggregation_functions import WeightedAverage

logger = logging.getLogger(__name__)


class Assigner:
    """Custom assigner class.

    Attributes:
        agg_functions_by_task (dict): Dictionary mapping tasks to their
            respective aggregation functions.
        agg_functions_by_task_name (dict): Dictionary mapping task names to
            their respective aggregation functions.
        authorized_cols (list of str): List of authorized collaborators.
        rounds_to_train (int): Number of rounds to train.
        all_tasks_for_round (defaultdict): Dictionary mapping round numbers
            to tasks.
        collaborators_for_task (defaultdict): Dictionary mapping round numbers
            to collaborators for each task.
        collaborator_tasks (defaultdict): Dictionary mapping round numbers
            to tasks for each collaborator.
        assigner_function (function): Function to assign tasks to
            collaborators.
    """

    def __init__(
        self, *, assigner_function, aggregation_functions_by_task, authorized_cols, rounds_to_train
    ):
        """Initialize the Custom assigner object.

        Args:
            assigner_function (function): Function to assign tasks to
                collaborators.
            aggregation_functions_by_task (dict): Dictionary mapping tasks to
                their respective aggregation functions.
            authorized_cols (list of str): List of authorized collaborators.
            rounds_to_train (int): Number of rounds to train.
        """
        self.agg_functions_by_task = aggregation_functions_by_task
        self.agg_functions_by_task_name = {}
        self.authorized_cols = authorized_cols
        self.rounds_to_train = rounds_to_train
        self.all_tasks_for_round = defaultdict(dict)
        self.collaborators_for_task = defaultdict(lambda: defaultdict(list))
        self.collaborator_tasks = defaultdict(lambda: defaultdict(list))
        self.assigner_function = assigner_function

        self.define_task_assignments()

    def define_task_assignments(self):
        """Define task assignments for each round and collaborator.

        This method uses the assigner function to assign tasks to
        collaborators for each round. It also maps tasks to their respective
        aggregation functions.

        Abstract method.

        Args:
            None

        Returns:
            None
        """
        for round_number in range(self.rounds_to_train):
            tasks_by_collaborator = self.assigner_function(
                self.authorized_cols,
                round_number,
                number_of_callaborators=len(self.authorized_cols),
            )
            for collaborator_name, tasks in tasks_by_collaborator.items():
                self.collaborator_tasks[round_number][collaborator_name].extend(tasks)
                for task in tasks:
                    self.all_tasks_for_round[round_number][task.name] = task
                    self.collaborators_for_task[round_number][task.name].append(collaborator_name)
                    if self.agg_functions_by_task:
                        self.agg_functions_by_task_name[task.name] = self.agg_functions_by_task.get(
                            task.function_name, WeightedAverage()
                        )

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Get tasks for a specific collaborator in a specific round.

        Abstract method.

        Args:
            collaborator_name (str): Name of the collaborator.
            round_number (int): Round number.

        Returns:
            list: List of tasks for the collaborator in the specified round.
        """
        return self.collaborator_tasks[round_number][collaborator_name]

    def get_collaborators_for_task(self, task_name, round_number):
        """Get collaborators for a specific task in a specific round.

        Abstract method.

        Args:
            task_name (str): Name of the task.
            round_number (int): Round number.

        Returns:
            list: List of collaborators for the task in the specified round.
        """
        return self.collaborators_for_task[round_number][task_name]

    def get_all_tasks_for_round(self, round_number):
        """Get all tasks for a specific round.

        Currently all tasks are performed on each round,
        But there may be a reason to change this.

        Args:
            round_number (int): Round number.

        Returns:
            list: List of all tasks for the specified round.
        """
        return [task.name for task in self.all_tasks_for_round[round_number].values()]

    def get_aggregation_type_for_task(self, task_name):
        """Get the aggregation type for a specific task (from self.tasks).

        Args:
            task_name (str): Name of the task.

        Returns:
            function: Aggregation function for the task.
        """
        agg_fn = self.agg_functions_by_task_name.get(task_name, WeightedAverage())
        return agg_fn



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""TensorDB Module."""

from threading import Lock
from types import MethodType
from typing import Dict, Iterator, Optional

import numpy as np
import pandas as pd

from openfl.databases.utilities import ROUND_PLACEHOLDER, _retrieve, _search, _store
from openfl.interface.aggregation_functions import AggregationFunction
from openfl.utilities import LocalTensor, TensorKey, change_tags


class TensorDB:
    """The TensorDB stores a tensor key and the data that it corresponds to.

    It is built on top of a pandas dataframe for it's easy insertion, retrieval
    and aggregation capabilities. Each collaborator and aggregator has its own
    TensorDB.

    Attributes:
        tensor_db: A pandas DataFrame that stores the tensor key and the data
            that it corresponds to.
        mutex: A threading Lock object used to ensure thread-safe operations
            on the tensor_db Dataframe.
    """

    def __init__(self) -> None:
        """Initializes a new instance of the TensorDB class."""
        types_dict = {
            "tensor_name": "string",
            "origin": "string",
            "round": "int32",
            "report": "bool",
            "tags": "object",
            "nparray": "object",
        }
        self.tensor_db = pd.DataFrame(
            {col: pd.Series(dtype=dtype) for col, dtype in types_dict.items()}
        )
        self._bind_convenience_methods()

        self.mutex = Lock()

    def _bind_convenience_methods(self):
        """Bind convenience methods for the TensorDB dataframe to make storage,
        retrieval, and search easier."""
        if not hasattr(self.tensor_db, "store"):
            self.tensor_db.store = MethodType(_store, self.tensor_db)
        if not hasattr(self.tensor_db, "retrieve"):
            self.tensor_db.retrieve = MethodType(_retrieve, self.tensor_db)
        if not hasattr(self.tensor_db, "search"):
            self.tensor_db.search = MethodType(_search, self.tensor_db)

    def __repr__(self) -> str:
        """Returns the string representation of the TensorDB object.

        Returns:
            content (str): The string representation of the TensorDB object.
        """
        with pd.option_context("display.max_rows", None):
            content = self.tensor_db[["tensor_name", "origin", "round", "report", "tags"]]
            return f"TensorDB contents:\n{content}"

    def __str__(self) -> str:
        """Returns the string representation of the TensorDB object.

        Returns:
        __repr__ (str): The string representation of the TensorDB object.
        """
        return self.__repr__()

    def clean_up(self, remove_older_than: int = 1) -> None:
        """Removes old entries from the database to prevent it from becoming
        too large and slow.

        Args:
            remove_older_than (int, optional): Entries older than this number
                of rounds are removed. Defaults to 1.
        """
        if remove_older_than < 0:
            # Getting a negative argument calls off cleaning
            return
        current_round = self.tensor_db["round"].astype(int).max()
        if current_round == ROUND_PLACEHOLDER:
            current_round = np.sort(self.tensor_db["round"].astype(int).unique())[-2]
        self.tensor_db = self.tensor_db[
            (self.tensor_db["round"].astype(int) > current_round - remove_older_than)
            | self.tensor_db["report"]
        ].reset_index(drop=True)

    def cache_tensor(self, tensor_key_dict: Dict[TensorKey, np.ndarray]) -> None:
        """Insert a tensor into TensorDB (dataframe).

        Args:
            tensor_key_dict (Dict[TensorKey, np.ndarray]): A dictionary where
                the key is a TensorKey and the value is a numpy array.

        Returns:
            None
        """
        entries_to_add = []
        with self.mutex:
            for tensor_key, nparray in tensor_key_dict.items():
                tensor_name, origin, fl_round, report, tags = tensor_key
                entries_to_add.append(
                    pd.DataFrame(
                        [
                            [
                                tensor_name,
                                origin,
                                fl_round,
                                report,
                                tags,
                                nparray,
                            ]
                        ],
                        columns=list(self.tensor_db.columns),
                    )
                )

            self.tensor_db = pd.concat([self.tensor_db, *entries_to_add], ignore_index=True)

    def get_tensor_from_cache(self, tensor_key: TensorKey) -> Optional[np.ndarray]:
        """Perform a lookup of the tensor_key in the TensorDB.

        Args:
            tensor_key (TensorKey): The key of the tensor to look up.

        Returns:
            Optional[np.ndarray]: The numpy array if it is available.
                Otherwise, returns None.
        """
        tensor_name, origin, fl_round, report, tags = tensor_key

        # TODO come up with easy way to ignore compression
        df = self.tensor_db[
            (self.tensor_db["tensor_name"] == tensor_name)
            & (self.tensor_db["origin"] == origin)
            & (self.tensor_db["round"] == fl_round)
            & (self.tensor_db["report"] == report)
            & (self.tensor_db["tags"] == tags)
        ]

        if len(df) == 0:
            return None
        return np.asarray(df["nparray"].iloc[0])

    def get_tensors_by_filter(self, custom_filter=None) -> dict:
        """Retrieve all tensors that match the specified round and tags.

        Args:
            custom_filter (callable): A function that takes a DataFrame and
                returns a boolean mask to filter the DataFrame.
                If None, no filtering is applied.
        Returns:
            dict: A dictionary where the keys are TensorKey objects and the values are numpy arrays.
        """
        # Construct a dictionary mapping TensorKey to np.ndarray
        tensor_dict = {}
        for _, row in self.tensor_db.iterrows():
            if custom_filter is not None:
                # Apply the custom filter to the DataFrame and skip rows that do not match
                if not custom_filter(row.to_dict()):
                    continue
            # Create a TensorKey object from the row data
            tensor_key = TensorKey(
                tensor_name=row["tensor_name"],
                origin=row["origin"],
                round_number=row["round"],
                report=row["report"],
                tags=row["tags"],
            )
            tensor_dict[tensor_key] = np.array(row["nparray"])

        return tensor_dict

    def get_tensors_by_round_and_tags(
        self,
        fl_round: int,
        tags: tuple,
    ) -> dict:
        """Retrieve all tensors that match the specified round and tags.

        Args:
            fl_round (int): The round number to filter tensors.
            tags (tuple): The tags to filter tensors.

        Returns:
            dict: A dictionary where the keys are TensorKey objects and the values are numpy arrays.
        """
        return self.get_tensors_by_filter(
            custom_filter=lambda row: row["round"] == fl_round and tuple(row["tags"]) == tuple(tags)
        )

    def get_aggregated_tensor(
        self,
        tensor_key: TensorKey,
        collaborator_weight_dict: dict,
        aggregation_function: AggregationFunction,
    ) -> Optional[np.ndarray]:
        """
        Determine whether all of the collaborator tensors are present for a
        given tensor key

        Returns their weighted average.

        Args:
            tensor_key (TensorKey): The tensor key to be resolved. If origin
                'agg_uuid' is present, can be returned directly. Otherwise
                must compute weighted average of all collaborators.
            collaborator_weight_dict (dict): A dictionary where the keys are
                collaborator names and the values are their respective weights.
            aggregation_function (AggregationFunction): Call the underlying
                numpy aggregation function to use to compute the weighted
                average. Default is just the weighted average.

        Returns:
            agg_nparray Optional[np.ndarray]: weighted_nparray The weighted
                average if all collaborator values are present. Otherwise,
                returns None.
            None: if not all values are present.
        """
        if len(collaborator_weight_dict) != 0:
            assert np.abs(1.0 - sum(collaborator_weight_dict.values())) < 0.01, (
                f"Collaborator weights do not sum to 1.0: {collaborator_weight_dict}"
            )

        collaborator_names = collaborator_weight_dict.keys()
        agg_tensor_dict = {}

        # Check if the aggregated tensor is already present in TensorDB
        tensor_name, origin, fl_round, report, tags = tensor_key

        raw_df = self.tensor_db[
            (self.tensor_db["tensor_name"] == tensor_name)
            & (self.tensor_db["origin"] == origin)
            & (self.tensor_db["round"] == fl_round)
            & (self.tensor_db["report"] == report)
            & (self.tensor_db["tags"] == tags)
        ]["nparray"]
        if len(raw_df) > 0:
            return np.asarray(raw_df.iloc[0]), {}

        for col in collaborator_names:
            new_tags = change_tags(tags, add_field=col)
            raw_df = self.tensor_db[
                (self.tensor_db["tensor_name"] == tensor_name)
                & (self.tensor_db["origin"] == origin)
                & (self.tensor_db["round"] == fl_round)
                & (self.tensor_db["report"] == report)
                & (self.tensor_db["tags"] == new_tags)
            ]["nparray"]
            if len(raw_df) == 0:
                tk = TensorKey(tensor_name, origin, report, fl_round, new_tags)
                print(f"No results for collaborator {col}, TensorKey={tk}")
                return None
            else:
                agg_tensor_dict[col] = raw_df.iloc[0]

        local_tensors = [
            LocalTensor(
                col_name=col_name,
                tensor=agg_tensor_dict[col_name],
                weight=collaborator_weight_dict[col_name],
            )
            for col_name in collaborator_names
        ]

        if hasattr(aggregation_function, "_privileged"):
            if aggregation_function._privileged:
                with self.mutex:
                    self._bind_convenience_methods()
                    agg_nparray = aggregation_function(
                        local_tensors,
                        self.tensor_db,
                        tensor_name,
                        fl_round,
                        tags,
                    )
                self.cache_tensor({tensor_key: agg_nparray})

                return np.array(agg_nparray)

        db_iterator = self._iterate()
        agg_nparray = aggregation_function(local_tensors, db_iterator, tensor_name, fl_round, tags)
        self.cache_tensor({tensor_key: agg_nparray})

        return np.asarray(agg_nparray)

    def _iterate(self, order_by: str = "round", ascending: bool = False) -> Iterator[pd.Series]:
        """Returns an iterator over the rows of the TensorDB, sorted by a
        specified column.

        Args:
            order_by (str, optional): The column to sort by. Defaults to
                'round'.
            ascending (bool, optional): Whether to sort in ascending order.
                Defaults to False.

        Returns:
            Iterator[pd.Series]: An iterator over the rows of the TensorDB.
        """
        columns = ["round", "nparray", "tensor_name", "tags"]
        rows = self.tensor_db[columns].sort_values(by=order_by, ascending=ascending).iterrows()
        for _, row in rows:
            yield row



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from openfl.transport.grpc import AggregatorGRPCClient, AggregatorGRPCServer
from openfl.transport.rest import AggregatorRESTClient, AggregatorRESTServer



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
from openfl.callbacks.callback import Callback
from openfl.callbacks.callback_list import CallbackList
from openfl.callbacks.lambda_callback import LambdaCallback
from openfl.callbacks.memory_profiler import MemoryProfiler
from openfl.callbacks.metric_writer import MetricWriter
from openfl.callbacks.secure_aggregation import SecAggBootstrapping



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""OpenFL Connector Module."""

from importlib import util

if util.find_spec("flwr") is not None:
    from openfl.component.connector.connector_flower import ConnectorFlower



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from importlib import util

if util.find_spec("flwr") is not None:
    from openfl.transport.grpc.interop.flower.interop_client import FlowerInteropClient
    from openfl.transport.grpc.interop.flower.interop_server import FlowerInteropServer



# Copyright (C) 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from openfl.federated.data.loader import DataLoader
import pandas as pd
import os
import subprocess


class SmokersHealthDataLoader(DataLoader):
    """Data Loader for Smokers Health Dataset."""

    def __init__(self, batch_size, data_path, **kwargs):
        super().__init__(**kwargs)

        # If data_path is None, this is being used for model initialization only
        if data_path is None:
            return

        # Load actual data if a data path is provided
        try:
            int(data_path)
        except ValueError:
            raise ValueError(
                f"Expected '{data_path}' to be representable as `int`, "
                "as it refers to the data shard number used by the collaborator."
            )

        # Download and prepare data
        self._download_raw_data()
        self.data_shard = self.load_data_shard(
            shard_num=int(data_path), **kwargs
        )

    def _download_raw_data(self):
        """
        Downloads and extracts the raw data for the smokers' health dataset.
        This method performs the following steps:
        1. Downloads the dataset from the specified Kaggle URL using the `curl` command.
        2. Saves the downloaded file as a ZIP archive in the `./data` directory.
        3. Extracts the contents of the ZIP archive into the `data` directory.
        """

        download_path = os.path.expanduser('./data/smokers_health.zip')
        subprocess.run(
            [
                'curl', '-L', '-o', download_path,
                'https://www.kaggle.com/api/v1/datasets/download/jaceprater/smokers-health-data'
            ],
            check=True
        )

        # Unzip the downloaded file into the data directory
        subprocess.run(['unzip', '-o', download_path, '-d', 'data'], check=True)

    def load_data_shard(self, shard_num, **kwargs):
        """
        Loads data from a CSV file.
        This method reads the data from a CSV file located at './data/smoking_health_data_final.csv'
        and returns it as a pandas DataFrame.
        Returns:
            pd.DataFrame: The data loaded from the CSV file.
        """
        file_path = os.path.join('data', 'smoking_health_data_final.csv')
        df = pd.read_csv(file_path)

        # Split data into shards
        shard_size = len(df) // shard_num
        start_idx = shard_size * (shard_num - 1)
        end_idx = start_idx + shard_size

        return df.iloc[start_idx:end_idx]

    def query(self, columns, **kwargs):
        """
        Query the data shard for the specified columns.
        Args:
            columns (list): A list of column names to query from the data shard.
            **kwargs: Additional keyword arguments (currently not used).
        Returns:
            DataFrame: A DataFrame containing the data for the specified columns.
        Raises:
            ValueError: If the columns parameter is not a list.
        """
        if not isinstance(columns, list):
            raise ValueError("Columns parameter must be a list")
        return self.data_shard[columns]

    def get_feature_shape(self):
        """
        This function is not required and is kept for compatibility.

        Returns:
            None
        """
        pass



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Workspace utils module."""

import logging
import os
import shutil
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from subprocess import check_call  # nosec
from sys import executable
from typing import Optional, Tuple, Union

from pip._internal.operations import freeze

logger = logging.getLogger(__name__)


class ExperimentWorkspace:
    """
    Experiment workspace context manager.

    This class is a context manager for creating a workspace for an experiment.

    Attributes:
        experiment_name (str): The name of the experiment.
        data_file_path (Path): The path to the data file for the experiment.
        install_requirements (bool): Whether to install the requirements for
            the experiment.
        cwd (Path): The current working directory.
        experiment_work_dir (Path): The working directory for the experiment.
        remove_archive (bool): Whether to remove the archive after the
            experiment.
    """

    def __init__(
        self,
        experiment_name: str,
        data_file_path: Path,
        install_requirements: bool = False,
        remove_archive: bool = True,
    ) -> None:
        """
        Initialize workspace context manager.

        Args:
            experiment_name (str): The name of the experiment.
            data_file_path (Path): The path to the data file for the
                experiment.
            install_requirements (bool, optional): Whether to install the
                requirements for the experiment. Defaults to False.
            remove_archive (bool, optional): Whether to remove the archive
                after the experiment. Defaults to True.
        """
        self.experiment_name = experiment_name
        self.data_file_path = data_file_path
        self.install_requirements = install_requirements
        self.cwd = Path.cwd()
        self.experiment_work_dir = self.cwd / self.experiment_name
        self.remove_archive = remove_archive

    def _install_requirements(self):
        """Install experiment requirements."""
        requirements_filename = self.experiment_work_dir / "requirements.txt"

        if requirements_filename.is_file():
            attempts = 10
            for _ in range(attempts):
                try:
                    check_call(
                        [
                            executable,
                            "-m",
                            "pip",
                            "install",
                            "-r",
                            requirements_filename,
                        ],
                        shell=False,
                    )
                except Exception as exc:
                    logger.error("Failed to install requirements: %s", exc)
                    # It's a workaround for cases when collaborators run
                    # in common virtual environment
                    time.sleep(5)
                else:
                    break
        else:
            logger.error("No " + requirements_filename + " file found.")

    def __enter__(self):
        """Create a collaborator workspace for the experiment."""
        if self.experiment_work_dir.exists():
            shutil.rmtree(self.experiment_work_dir, ignore_errors=True)
        os.makedirs(self.experiment_work_dir)

        shutil.unpack_archive(self.data_file_path, self.experiment_work_dir, format="zip")

        if self.install_requirements:
            self._install_requirements()

        os.chdir(self.experiment_work_dir)

        # This is needed for python module finder
        sys.path.append(str(self.experiment_work_dir))

    def __exit__(self, exc_type, exc_value, traceback):
        """Remove the workspace."""
        os.chdir(self.cwd)
        shutil.rmtree(self.experiment_work_dir, ignore_errors=True)
        if str(self.experiment_work_dir) in sys.path:
            sys.path.remove(str(self.experiment_work_dir))

        if self.remove_archive:
            logger.debug(
                f"Exiting from the workspace context manager for {self.experiment_name} experiment"
            )
            logger.debug(
                f"Exiting from the workspace context manager for {self.experiment_name} experiment"
            )
            logger.debug("Archive still exists: %s", self.data_file_path.exists())
            self.data_file_path.unlink(missing_ok=False)


def dump_requirements_file(
    path: Union[str, Path] = "./requirements.txt",
    keep_original_prefixes: bool = True,
    prefixes: Optional[Union[Tuple[str], str]] = None,
) -> None:
    """
    Prepare and save requirements.txt.

    This function prepares a requirements.txt file and saves it to a specified
    path.

    Args:
        path (Union[str, Path], optional): The path to save the
            requirements.txt file.
            Defaults to './requirements.txt'.
        keep_original_prefixes (bool, optional): Whether to keep the original
            prefixes in the requirements.txt file. Defaults to True.
        prefixes (Optional[Union[Tuple[str], str]], optional): The prefixes to
            add to the requirements.txt file. Defaults to None.
    """
    path = Path(path).absolute()

    # Prepare user provided prefixes for merge with original ones
    if prefixes is None:
        prefixes = set()
    elif type(prefixes) is str:
        prefixes = set(
            prefixes,
        )
    else:
        prefixes = set(prefixes)

    # Merge prefixes:
    # We expect that all the prefixes in a requirement file
    # are placed at the top
    if keep_original_prefixes and path.is_file():
        with open(path, encoding="utf-8") as f:
            for line in f:
                if line == "\n":
                    continue
                if line[0] == "-":
                    prefixes |= {line.replace("\n", "")}
                else:
                    break

    requirements_generator = freeze.freeze()
    with open(path, "w", encoding="utf-8") as f:
        for prefix in prefixes:
            f.write(prefix + "\n")

        for package in requirements_generator:
            if _is_package_versioned(package):
                f.write(package + "\n")


def _is_package_versioned(package: str) -> bool:
    """
    Check if a package has a version.

    Args:
        package (str): The package to check.

    Returns:
        bool: `True` if the package has a version, `False` otherwise.
    """
    return (
        "==" in package
        and package not in ["pkg-resources==0.0.0", "pkg_resources==0.0.0"]
        and "-e " not in package
    )


@contextmanager
def set_directory(path: Path):
    """Set the current working directory within the context.

    Args:
        path (Path): The path to set as the current working directory.

    Yields:
        None
    """

    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import os

import nibabel as nib
import numpy as np
import numpy.ma as ma


def parse_segments(seg, msk_modes):
    """Parse the label segments.

    Each channel corresponds to a different region of the tumor, decouple and stack these

    mode_to_key_value = {"necrotic": 1, "edema": 2, "GD": 4}

    Args:
        seg: The segmentation labels
        msk_modes: Label mode to parse for the model

    Returns:
        The processed mask labels

    """
    msks_parsed = []
    for slice_ in range(seg.shape[-1]):
        # which mask values indicicate which label mode
        mode_to_key_value = {'necrotic': 1, 'edema': 2, 'GD': 4}
        curr = seg[:, :, slice_]
        this_msk_parts = []
        for mode in msk_modes:
            this_msk_parts.append(
                ma.masked_not_equal(curr, mode_to_key_value[mode]).filled(
                    fill_value=0
                )
            )
        msks_parsed.append(np.dstack(this_msk_parts))

    # Replace all tumorous areas with 1 (previously marked as 1, 2 or 4)
    mask = np.asarray(msks_parsed)
    mask[mask > 0] = 1

    return mask


def normalize_stack(imgs):
    """Z-score normalization of the input images.

    Args:
        imgs: The input images

    Returns:
        The processed (normalized) input images

    """
    imgs = (imgs - np.mean(imgs)) / (np.std(imgs))
    return imgs


def resize_data(dataset, new_size=128, rotate=3):
    """Resize 2D images within data by cropping equally from their boundaries.

    Args:
        dataset (np.array): Data containing 2D images whose dimensions are
        along the 1st and 2nd axes.
        new_size (int): Dimensions of square image to which resizing will
        occur. Assumed to be an even distance away from both dimensions of
        the images within dataset. (default 128) rotate (int): Number of
        counter clockwise 90 degree rotations to perform.

    Returns:
        (np.array): resized data

    Raises:
        ValueError: If (dataset.shape[1] - new_size) and
         (dataset.shape[2] - new_size) are not both even integers.

    """
    # Determine whether dataset and new_size are compatible with existing logic
    if (dataset.shape[1] - new_size) % 2 != 0 and (dataset.shape[2] - new_size) % 2 != 0:
        raise ValueError(f'dataset shape: {dataset.shape} and new_size: {new_size} '
                         f'are not compatible with existing logic')

    start_index = int((dataset.shape[1] - new_size) / 2)
    end_index = dataset.shape[1] - start_index

    if rotate != 0:
        resized = np.rot90(dataset[:, start_index:end_index, start_index:end_index],
                           rotate, axes=(1, 2))
    else:
        resized = dataset[:, start_index:end_index, start_index:end_index]

    return resized


# adapted from https://github.com/NervanaSystems/topologies
def _update_channels(imgs, msks, img_channels_to_keep,
                     msk_channels_to_keep, channels_last):
    """Filter the channels of images and move placement of channels in shape if desired.

    Args:
        imgs (np.array): A stack of images with channels (channels could be
        anywhere in number from one to four. Images are indexed along first
        axis, with channels along fourth (last) axis. msks (np.array): A stack
        of binary masks with channels (channels could be anywhere in number
        from one to four. Images are indexed along first axis, with channels
        along fourth (last) axis. img_channels_to_keep (flat np.ndarray): the
        channels to keep in the image (remove others) msk_channels_to_keep
        (flat np.ndarray): the channels to sum in the mask (resulting in
        a single channel array)
        channels_last (bool): Return channels in last axis? otherwise just
         after first

    Returns:
        images, masks with selected channels
    """
    new_imgs = imgs[:, :, :, img_channels_to_keep]
    # the mask channels that are kept are summed over to leave one channel
    # note the indices producing non-zero entries on these masks are mutually exclusive
    # so that the result continues to be an array with only ones and zeros
    msk_summands = [
        msks[:, :, :, channel:channel + 1] for channel in msk_channels_to_keep
    ]
    new_msks = np.sum(msk_summands, axis=0)

    if not channels_last:
        new_order = [0, 3, 1, 2]
        return np.transpose(new_imgs, new_order), np.transpose(new_msks, new_order)
    else:
        return new_imgs, new_msks


def list_files(root, extension, parts):
    """Construct files from root, parts."""
    files = [root + part + extension for part in parts]
    return files


def nii_reader(brain_path, task, channels_last=True,
               numpy_type='float64', normalization='by_mode', **kwargs):
    """Fetch a whole brain 3D image from disc.

    Assumes data_dir contains only subdirectories, each containing exactly one
    of each of the following files: "<subdir>_<type>.nii.gz", where <subdir> is
    the name of the subdirectory and <type> is one of ["t1", "t2","flair","t1ce"],
    as well as a segmentation label file "<subdir>_<suffix>", where suffix is:
    'seg_binary.nii.gz', 'seg_binarized.nii.gz', or 'SegBinarized.nii.gz'.
    These files provide all modes and segmentation label for a single patient
    brain scan consisting of 155 axial slice images. The reader returns the whole
    brain (all 155 slices).
    Note: Much of the logic here is included in order to allow for the reader
    to have the same functionality (shared code) regardless of normalization
    method. This allows us to validate all functionality except the normalization,
    against data on disc that was previously processed by normalizing all scanning
    modalities together. Post validation as desired, the code can be simplified.

    Args:
        brain_path (str): path to files containing image features and label
        set for a single brain.
        task (string): Describes the classification task, which determines
        the way in which the scanning modes and label information is filtered
        and combined.
        channels_last (bool): Determines whether reader should output channels
        in the position of the last axis.Otherwise channels are placed just
        after first axis.
        numpy_type (string): The numpy datatype for final casting before return.
        normalization (string): Determines whether the feature scanning modes
        are not normalized (None), normalized by scanning mode ('by_mode'),
        or normalized across all modes ('modes_together').
        **kwargs: unused

    Returns:
        np.array: whole 3D brain associated to a subdirectory given by the index

    Raises:
        ValueError: If label_type is not in
                    ['whole_tumor', 'enhanced_tumor', 'active_core', 'other']
        ValueError: If the path determined by idx and indexed_data_paths points
                    to a file with incomplete data

    """
    files = os.listdir(brain_path)
    # link task to appropriate image and mask channels of interest
    img_modes = ['t1', 't2', 'flair', 't1ce']
    msk_modes = ['necrotic', 'edema', 'GD']
    task_to_img_modes = {
        'whole_tumor': ['flair'],
        'enhanced_tumor': ['t1'],
        'active_core': ['t2'],
        'other': ['t1', 't2', 'flair', 't1ce'],
    }
    task_to_msk_modes = {
        'whole_tumor': ['necrotic', 'edema', 'GD'],
        'enhanced_tumor': ['GD'],
        'active_core': ['edema', 'GD'],
        'other': ['necrotic', 'edema', 'GD'],
    }
    msk_names = ['seg_binary', 'seg_binarized', 'SegBinarized', 'seg']

    # validate that task is an allowed key
    if task not in task_to_img_modes.keys():
        raise ValueError(f'{task} is not a valid task')

    # validate that the tasks used in task_to_img_modes and
    # task_to_msk_modes are the same
    if set(task_to_img_modes.keys()) != set(task_to_msk_modes.keys()):
        raise RuntimeError('Hard coded keys to task_to_img_modes'
                           'and task_to_mask_modes are not the same and should be.')

    # check that all appropriate files are present
    file_root = brain_path.split('/')[-1] + '_'
    # check for all possible extensions
    extensions = ['.nii.gz', '.nii']

    # record files needed
    # needed mask files are currently independent of task
    need_files_oneof = None
    need_files_all = None
    for extension in extensions:
        need_files_oneof = list_files(file_root, extension, msk_names)
        if normalization != 'modes_together':
            need_files_all = list_files(file_root, extension, task_to_img_modes[task])
        else:
            need_files_all = list_files(file_root, extension, img_modes)

        correct_files = np.all([
            (reqd in files)
            for reqd in need_files_all
        ]) and np.sum([
            (reqd in files)
            for reqd in need_files_oneof
        ]) == 1
        if correct_files:
            break
    if not correct_files:
        return None, None

    # get image (features)
    imgs_per_mode = []
    for file in need_files_all:
        path = os.path.join(brain_path, file)
        full_brain = np.array(nib.load(path).dataobj)
        imgs_per_mode.append(resize_data(np.transpose(full_brain, [-1, 0, 1])))

    # normalize each model then stack, stack after normalizing all modes together,
    # or stack without normalizing at all
    if normalization == 'by_mode':
        imgs = np.stack([normalize_stack(imgs) for imgs in imgs_per_mode], axis=-1)
    elif normalization == 'modes_together':
        imgs = normalize_stack(np.stack(imgs_per_mode, axis=-1))
    elif normalization is None:
        imgs = np.stack(imgs_per_mode, axis=-1)
    else:
        raise ValueError(f'{normalization} is not a supported normalization.')

    # get mask (labels)
    for file in need_files_oneof:
        if file in files:
            path = os.path.join(brain_path, file)
            full_brain_msk = np.array(nib.load(path).dataobj)
            msks = resize_data(parse_segments(full_brain_msk, msk_modes))
            break

    # determine which channels are wanted in our images in the case where we kept
    # extra scanning modes in order to normalize across all but only use a subset
    msk_mode_to_channel = {
        mode: channel_num for (channel_num, mode) in enumerate(msk_modes)
    }
    msk_channels_to_keep = np.array(
        [msk_mode_to_channel[mode] for mode in task_to_msk_modes[task]])
    # if we normalization is by_mode or None,
    # we have already restricted the image channels
    if normalization in ['by_mode', None]:
        img_channels_to_keep = np.arange(imgs.shape[-1])
    elif normalization == 'modes_together':
        img_mode_to_channel = {
            mode: channel_num for (channel_num, mode) in enumerate(img_modes)
        }
        img_channels_to_keep = np.array(
            [img_mode_to_channel[mode] for mode in task_to_img_modes[task]]
        )
    else:
        raise ValueError(f'{normalization} is not a supported normalization.')

    img = imgs
    msk = msks

    img, msk = _update_channels(
        img, msk, img_channels_to_keep, msk_channels_to_keep, channels_last)

    return img.astype(numpy_type), msk.astype(numpy_type)



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from importlib import util

from openfl.pipelines.kc_pipeline import KCPipeline
from openfl.pipelines.no_compression_pipeline import NoCompressionPipeline
from openfl.pipelines.random_shift_pipeline import RandomShiftPipeline
from openfl.pipelines.skc_pipeline import SKCPipeline
from openfl.pipelines.stc_pipeline import STCPipeline
from openfl.pipelines.tensor_codec import TensorCodec

if util.find_spec("torch") is not None:
    from openfl.pipelines.eden_pipeline import EdenPipeline  # NOQA



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from openfl.protocols import aggregator_pb2 as openfl_dot_protocols_dot_aggregator__pb2
from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2

GRPC_GENERATED_VERSION = '1.65.5'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in openfl/protocols/aggregator_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class AggregatorStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Ping = channel.unary_unary(
                '/openfl.aggregator.Aggregator/Ping',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.PingRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.PingResponse.FromString,
                _registered_method=True)
        self.GetTasks = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetTasks',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.FromString,
                _registered_method=True)
        self.GetAggregatedTensor = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetAggregatedTensor',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorResponse.FromString,
                _registered_method=True)
        self.SendLocalTaskResults = channel.stream_unary(
                '/openfl.aggregator.Aggregator/SendLocalTaskResults',
                request_serializer=openfl_dot_protocols_dot_base__pb2.DataStream.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.FromString,
                _registered_method=True)
        self.InteropRelay = channel.unary_unary(
                '/openfl.aggregator.Aggregator/InteropRelay',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.FromString,
                _registered_method=True)


class AggregatorServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTasks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAggregatedTensor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendLocalTaskResults(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InteropRelay(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AggregatorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.PingRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.PingResponse.SerializeToString,
            ),
            'GetTasks': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTasks,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.SerializeToString,
            ),
            'GetAggregatedTensor': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAggregatedTensor,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorResponse.SerializeToString,
            ),
            'SendLocalTaskResults': grpc.stream_unary_rpc_method_handler(
                    servicer.SendLocalTaskResults,
                    request_deserializer=openfl_dot_protocols_dot_base__pb2.DataStream.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.SerializeToString,
            ),
            'InteropRelay': grpc.unary_unary_rpc_method_handler(
                    servicer.InteropRelay,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'openfl.aggregator.Aggregator', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('openfl.aggregator.Aggregator', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Aggregator(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/Ping',
            openfl_dot_protocols_dot_aggregator__pb2.PingRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.PingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTasks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/GetTasks',
            openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAggregatedTensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/GetAggregatedTensor',
            openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendLocalTaskResults(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/openfl.aggregator.Aggregator/SendLocalTaskResults',
            openfl_dot_protocols_dot_base__pb2.DataStream.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InteropRelay(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/InteropRelay',
            openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import torch
import torchvision
import numpy as np
import pathlib
import os
import matplotlib
import matplotlib.pyplot as plt
import PIL.Image as Image
import imagen as ig
import numbergen as ng


watermark_dir = "./files/watermark-dataset/MWAFFLE/"


def generate_watermark(
    x_size=28, y_size=28, num_class=10, num_samples_per_class=10, img_dir=watermark_dir
):
    """
    Generate Watermark by superimposing a pattern on noisy background.

    Parameters
    ----------
    x_size: x dimension of the image
    y_size: y dimension of the image
    num_class: number of classes in the original dataset
    num_samples_per_class: number of samples to be generated per class
    img_dir: directory for saving watermark dataset

    Reference
    ---------
    WAFFLE: Watermarking in Federated Learning (https://arxiv.org/abs/2008.07298)

    """
    x_pattern = int(x_size * 2 / 3.0 - 1)
    y_pattern = int(y_size * 2 / 3.0 - 1)

    np.random.seed(0)
    for cls in range(num_class):
        patterns = []
        random_seed = 10 + cls
        patterns.append(
            ig.Line(
                xdensity=x_pattern,
                ydensity=y_pattern,
                thickness=0.001,
                orientation=np.pi * ng.UniformRandom(seed=random_seed),
                x=ng.UniformRandom(seed=random_seed) - 0.5,
                y=ng.UniformRandom(seed=random_seed) - 0.5,
                scale=0.8,
            )
        )
        patterns.append(
            ig.Arc(
                xdensity=x_pattern,
                ydensity=y_pattern,
                thickness=0.001,
                orientation=np.pi * ng.UniformRandom(seed=random_seed),
                x=ng.UniformRandom(seed=random_seed) - 0.5,
                y=ng.UniformRandom(seed=random_seed) - 0.5,
                size=0.33,
            )
        )

        pat = np.zeros((x_pattern, y_pattern))
        for i in range(6):
            j = np.random.randint(len(patterns))
            pat += patterns[j]()
        res = pat > 0.5
        pat = res.astype(int)

        x_offset = np.random.randint(x_size - x_pattern + 1)
        y_offset = np.random.randint(y_size - y_pattern + 1)

        for i in range(num_samples_per_class):
            base = np.random.rand(x_size, y_size)
            base[
                x_offset: x_offset + pat.shape[0],
                y_offset: y_offset + pat.shape[1],
            ] += pat
            d = np.ones((x_size, x_size))
            img = np.minimum(base, d)
            if not os.path.exists(img_dir + str(cls) + "/"):
                os.makedirs(img_dir + str(cls) + "/")
            plt.imsave(
                img_dir + str(cls) + "/wm_" + str(i + 1) + ".png",
                img,
                cmap=matplotlib.cm.gray,
            )


# If the Watermark dataset does not exist, generate and save the Watermark images
watermark_path = pathlib.Path(watermark_dir)
if watermark_path.exists() and watermark_path.is_dir():
    print(
        f"Watermark dataset already exists at: {watermark_path}. Proceeding to next step ... "
    )
    pass
else:
    print("Generating Watermark dataset... ")
    generate_watermark()


class WatermarkDataset(torch.utils.data.Dataset):
    def __init__(self, images_dir, label_dir=None, transforms=None):
        self.images_dir = os.path.abspath(images_dir)
        self.image_paths = [
            os.path.join(self.images_dir, d) for d in os.listdir(self.images_dir)
        ]
        self.label_paths = label_dir
        self.transform = transforms
        temp = []

        # Recursively counting total number of images in the directory
        for image_path in self.image_paths:
            for path in os.walk(image_path):
                if len(path) <= 1:
                    continue
                path = path[2]
                for im_n in [image_path + "/" + p for p in path]:
                    temp.append(im_n)
        self.image_paths = temp

        if len(self.image_paths) == 0:
            raise Exception(f"No file(s) found under {images_dir}")

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image_filepath = self.image_paths[idx]
        image = Image.open(image_filepath)
        image = image.convert("RGB")
        image = self.transform(image)
        label = int(image_filepath.split("/")[-2])

        return image, label


def get_watermark_transforms():
    return torchvision.transforms.Compose(
        [
            torchvision.transforms.Grayscale(),
            torchvision.transforms.Resize(28),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(mean=(0.5,), std=(0.5,)),  # Normalize
        ]
    )


watermark_data = WatermarkDataset(
    images_dir=watermark_dir,
    transforms=get_watermark_transforms(),
)


def aggregator_attrs(watermark_data, batch_size):
    return {
        "watermark_data_loader": torch.utils.data.DataLoader(
            watermark_data, batch_size=batch_size, shuffle=True
        ),
        "pretrain_epochs": 25,
        "retrain_epochs": 25,
        "watermark_acc_threshold": 0.98,
        "watermark_pretraining_completed": False,
    }



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.utilities.resources module."""

from logging import getLogger
from subprocess import PIPE, run

logger = getLogger(__name__)


def get_number_of_gpus() -> int:
    """Returns number of NVIDIA GPUs attached to the machine.

    This function executes the `nvidia-smi --list-gpus` command to get the
    list of GPUs.
    If the command fails (e.g., NVIDIA drivers are not installed), it logs a
    warning and returns 0.

    Returns:
        int: The number of NVIDIA GPUs attached to the machine.
    """
    # Execute the nvidia-smi command.
    command = "nvidia-smi --list-gpus"
    try:
        op = run(command.strip().split(), shell=False, stdout=PIPE, stderr=PIPE)
        stdout = op.stdout.decode().strip()
        return len(stdout.split("\n"))
    except FileNotFoundError:
        logger.warning(
            f'No GPUs found! If this is a mistake please try running "{command}" ' + "manually."
        )
        return 0



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Task package."""

from importlib import util

from openfl.federated.task.runner import TaskRunner  # NOQA

if util.find_spec("keras") is not None:
    from openfl.federated.task.runner_keras import KerasTaskRunner  # NOQA
if util.find_spec("torch") is not None:
    from openfl.federated.task.runner_pt import PyTorchTaskRunner  # NOQA
if util.find_spec("xgboost") is not None:
    from openfl.federated.task.runner_xgb import XGBoostTaskRunner  # NOQA
if util.find_spec("flwr") is not None:
    from openfl.federated.task.runner_flower import FlowerTaskRunner  # NOQA



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""AggregatorRESTClient module."""

# Standard library imports
import logging
import ssl
import struct
import time
from typing import Any, List, Tuple

# Third-party libraries
import requests
from google.protobuf import json_format
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

# Internal modules
from openfl.protocols import aggregator_pb2, base_pb2
from openfl.protocols.aggregator_client_interface import AggregatorClientInterface

logger = logging.getLogger(__name__)


class SecurityError(Exception):
    """Security-related error."""

    pass


class AggregatorRESTClient(AggregatorClientInterface):
    def __init__(
        self,
        agg_addr,
        agg_port,
        aggregator_uuid: str,
        federation_uuid: str,
        collaborator_name: str,
        use_tls=True,
        require_client_auth=True,
        root_certificate=None,
        certificate=None,
        private_key=None,
        single_col_cert_common_name=None,
        refetch_server_cert_callback=None,
        **kwargs,
    ):
        """
        Initialize the AggregatorRESTClient with proper security settings.

        Args:
            agg_addr: Aggregator address
            agg_port: Aggregator port
            aggregator_uuid: UUID of the aggregator
            federation_uuid: UUID of the federation
            collaborator_name: Name of the collaborator
            use_tls: Whether to use TLS
            require_client_auth: Whether to require client authentication
            root_certificate: Path to root certificate
            certificate: Path to client certificate
            private_key: Path to client private key
            single_col_cert_common_name: Common name for single collaborator certificate
            refetch_server_cert_callback: Callback to refetch server certificate
        """
        self.use_tls = use_tls
        self.require_client_auth = require_client_auth
        self.root_certificate = root_certificate
        self.certificate = certificate
        self.private_key = private_key
        self.aggregator_uuid = aggregator_uuid
        self.federation_uuid = federation_uuid
        self.collaborator_name = collaborator_name
        self.single_col_cert_common_name = single_col_cert_common_name
        self.refetch_server_cert_callback = refetch_server_cert_callback

        # Determine scheme and TLS verification
        scheme = "https" if self.use_tls else "http"

        # Configure certificate verification
        self.cert_verification = self._configure_cert_verification(
            self.use_tls, self.root_certificate
        )

        # Configure client certificates if required
        if self.use_tls and self.require_client_auth:
            if not self.certificate or not self.private_key:
                raise ValueError(
                    "Both certificate and private key are required for mTLS "
                    "(client authentication). "
                    "Please provide both certificate and private key paths."
                )
            self.cert = (self.certificate, self.private_key)
        else:
            self.cert = None

        # Configure session with proper settings
        self.session = requests.Session()

        # Set default headers
        self.session.headers.update(
            {
                "Connection": "keep-alive",
                "Keep-Alive": "timeout=300",
                "Accept": "application/json",
                "X-Content-Type-Options": "nosniff",
                "X-Frame-Options": "DENY",
                "X-XSS-Protection": "1; mode=block",
            }
        )

        # Configure timeouts with longer duration for large payloads
        self.timeout = (30, 300)  # (connect timeout, read timeout) in seconds

        # Configure retries with backoff
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[408, 429, 500, 502, 503, 504],
            allowed_methods=["GET", "POST"],
            raise_on_status=True,
        )

        # Configure the adapter with the retry strategy
        adapter = HTTPAdapter(
            max_retries=retry_strategy, pool_connections=10, pool_maxsize=10, pool_block=False
        )

        # Mount the adapter for both HTTP and HTTPS
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

        # Build the base URL
        self.base_url = f"{scheme}://{agg_addr}:{agg_port}/experimental/v1"

        # Log warning about experimental API
        logger.warning(
            "Initializing Aggregator REST Client (EXPERIMENTAL API - Not for production use)"
        )

        # Verify certificates if TLS is enabled
        if self.use_tls:
            try:
                self._verify_certificates()
            except Exception as e:
                logger.error(f"Certificate verification failed: {e}")
                raise

    @classmethod
    def _configure_cert_verification(
        cls, use_tls: bool, root_certificate: str = None
    ) -> bool | str:
        """
        Configure certificate verification settings for requests.

        Args:
            use_tls: Whether TLS is enabled
            root_certificate: Optional path to root certificate file

        Returns:
            Union[bool, str]: Either True for system CA bundle, False for no verification,
                             or path to root certificate file
        """
        if not use_tls:
            return False

        if root_certificate:
            return root_certificate

        return True  # Use system's default CA bundle

    def _verify_certificates(self):
        """Verify SSL certificates and configuration."""
        import socket
        import ssl

        # Try to establish a test connection
        try:
            hostname = self.base_url.split("://")[1].split(":")[0]
            port = int(self.base_url.split(":")[2].split("/")[0])

            # Create SSL context with specific options
            context = ssl.create_default_context()
            context.verify_mode = ssl.CERT_REQUIRED
            context.check_hostname = True

            # Set secure cipher suites
            context.set_ciphers("ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384")

            # Disable older TLS versions
            context.options |= (
                ssl.OP_NO_TLSv1
                | ssl.OP_NO_TLSv1_1
                | ssl.OP_NO_TLSv1_2
                | ssl.OP_NO_COMPRESSION
                | ssl.OP_NO_TICKET
            )

            if self.root_certificate:
                context.load_verify_locations(cafile=self.root_certificate)

            if self.certificate and self.private_key:
                context.load_cert_chain(certfile=self.certificate, keyfile=self.private_key)

            # Use context managers for proper resource cleanup
            with socket.create_connection((hostname, port)) as sock:
                with context.wrap_socket(sock, server_hostname=hostname) as _:
                    pass  # Connection successful if we get here

        except ssl.SSLError as e:
            if "CERTIFICATE_UNKNOWN" in str(e):
                logger.error(
                    "Certificate unknown error - this usually means the "
                    "server's certificate is not trusted"
                )
                logger.error("Please verify that:")
                logger.error(
                    "1. The root certificate contains all necessary intermediate certificates"
                )
                logger.error("2. The server's certificate is properly signed by a trusted CA")
                logger.error("3. The hostname matches the certificate's subject")
            raise
        except Exception as e:
            logger.error(f"Connection verification failed: {e}")
            raise

    def _build_header(self) -> dict:
        """Build and return a header dictionary with security headers."""
        headers = {
            "Receiver": self.aggregator_uuid,
            "Federation-UUID": self.federation_uuid,
            "Single-Col-Cert-CN": self.single_col_cert_common_name or "",
            "X-Content-Type-Options": "nosniff",
            "X-Frame-Options": "DENY",
            "X-XSS-Protection": "1; mode=block",
            "Sender": self.collaborator_name,
        }
        if self.use_tls:
            headers["Strict-Transport-Security"] = "max-age=31536000; includeSubDomains"
        return headers

    def _make_request(
        self, method, url, data=None, params=None, headers=None, stream=False, timeout=None
    ):
        """Make a request with proper security settings."""
        start_time = time.time()
        try:
            self._validate_url_scheme(url)
            request_headers = self._prepare_headers(headers)
            response = self._execute_request(
                method, url, request_headers, data, params, stream, timeout
            )
            self._validate_response(response)
            logger.debug(f"Request completed in {time.time() - start_time:.2f} seconds")
            return response

        except requests.exceptions.Timeout:
            logger.error(f"Request timed out after {time.time() - start_time:.2f} seconds")
            raise
        except requests.exceptions.ConnectionError as e:
            self._handle_connection_error(e)
            raise
        except requests.exceptions.RequestException as e:
            self._handle_request_error(e)
            raise

    def _validate_url_scheme(self, url):
        """Validate URL scheme matches TLS setting."""
        if self.use_tls and not url.startswith("https://"):
            raise ValueError("TLS required but URL is not HTTPS")
        elif not url.startswith("http://") and not url.startswith("https://"):
            raise ValueError("URL must use either HTTP or HTTPS scheme")

    def _prepare_headers(self, headers):
        """Prepare request headers with security settings."""
        request_headers = self._build_header()
        if headers:
            request_headers.update(headers)
        return request_headers

    def _execute_request(self, method, url, headers, data, params, stream, timeout):
        """Execute the HTTP request with retry logic."""
        max_retries = 3
        for attempt in range(max_retries):
            try:
                session = requests.Session()
                if self.use_tls:
                    # Extract hostname from URL for verification
                    hostname = url.split("://")[1].split(":")[0].split("/")[0]

                    # Create a custom SSL context for this request
                    context = ssl.create_default_context(
                        cafile=self.root_certificate if self.root_certificate else None
                    )
                    context.verify_mode = ssl.CERT_REQUIRED

                    # Configure session with SSL context and hostname verification
                    session.verify = self.cert_verification
                    session.cert = self.cert

                    # Configure adapter with proper SSL settings
                    adapter = HTTPAdapter(
                        pool_connections=1,
                        pool_maxsize=1,
                        max_retries=Retry(
                            total=3,
                            backoff_factor=1,
                            status_forcelist=[408, 429, 500, 502, 503, 504],
                            allowed_methods=["GET", "POST"],
                        ),
                    )
                    session.mount("https://", adapter)

                    # Build the complete headers with security information
                    base_headers = self._build_header()
                    if headers:
                        # Merge user-provided headers with base headers
                        base_headers.update(headers)
                    headers = base_headers
                    headers["Host"] = hostname

                    # Add certificate info to request kwargs
                    request_kwargs = {
                        "method": method,
                        "url": url,
                        "headers": headers,
                        "data": data,
                        "params": params,
                        "stream": stream,
                        "verify": self.cert_verification,
                        "timeout": timeout or self.timeout,
                    }

                    # Add client certificate if mTLS is enabled
                    if self.require_client_auth:
                        if not self.certificate or not self.private_key:
                            raise ValueError(
                                "Both certificate and private key are required for mTLS "
                                "(client authentication). "
                                "Please provide both certificate and private key paths."
                            )
                        # Use proper cert format
                        request_kwargs["cert"] = (self.certificate, self.private_key)

                    response = session.request(**request_kwargs)
                else:
                    # For non-TLS requests, still use the security headers
                    base_headers = self._build_header()
                    if headers:
                        base_headers.update(headers)

                    response = session.request(
                        method=method,
                        url=url,
                        headers=base_headers,
                        data=data,
                        params=params,
                        stream=stream,
                        timeout=timeout or self.timeout,
                    )
                return response
            except requests.exceptions.SSLError as e:
                self._handle_ssl_error(e, attempt, max_retries)
                if attempt == max_retries - 1:
                    raise

    def _handle_ssl_error(self, e, attempt, max_retries):
        """Handle SSL errors with retry logic."""
        if "CERTIFICATE_UNKNOWN" in str(e):
            logger.error(
                "Certificate unknown error - this usually means the "
                "server's certificate is not trusted"
            )
            logger.error("Please verify that:")
            logger.error("1. The root certificate contains all necessary intermediate certificates")
            logger.error("2. The server's certificate is properly signed by a trusted CA")
            logger.error("3. The hostname matches the certificate's subject")
            if attempt < max_retries - 1 and self.refetch_server_cert_callback:
                logger.debug("Attempting to refetch server certificate")
                self.root_certificate = self.refetch_server_cert_callback()
                # Update the cert_verification with the new root certificate
                self.cert_verification = self._configure_cert_verification(
                    self.use_tls, self.root_certificate
                )
                # Re-verify certificates
                try:
                    self._verify_certificates()
                except Exception as verify_error:
                    logger.error(f"Certificate re-verification failed: {verify_error}")
                    raise
            else:
                raise
        else:
            if attempt < max_retries - 1:
                logger.warning(f"SSL error (attempt {attempt + 1}/{max_retries}): {str(e)}")
                time.sleep(1)
            else:
                raise

    def _validate_response(self, response):
        """Validate response headers and security settings."""
        security_headers = {
            "X-Content-Type-Options": "nosniff",
            "X-Frame-Options": "DENY",
            "X-XSS-Protection": "1; mode=block",
        }
        for header, expected_value in security_headers.items():
            if header in response.headers and response.headers[header] != expected_value:
                logger.warning(f"Missing or incorrect security header: {header}")

        response.raise_for_status()

    def _handle_connection_error(self, e):
        """Handle connection errors."""
        logger.error(f"Connection error: {e}")
        if hasattr(e, "args") and len(e.args) > 0:
            logger.error(f"Connection error details: {e.args[0]}")

    def _handle_request_error(self, e):
        """Handle request errors."""
        logger.error(f"Request failed: {e}")
        if hasattr(e, "args") and len(e.args) > 0:
            logger.error(f"Request error details: {e.args[0]}")

    def get_tasks(self) -> Tuple[List[Any], int, int, bool]:
        """Get tasks from the aggregator with proper security settings."""
        headers = {"Accept": "application/json", "Sender": self.collaborator_name}
        params = {
            "collaborator_id": self.collaborator_name,
            "federation_uuid": self.federation_uuid,
        }
        url = f"{self.base_url}/tasks"
        response = self._make_request("GET", url, headers=headers, params=params)
        response.raise_for_status()
        data = response.json()
        tasks_resp = aggregator_pb2.GetTasksResponse()
        json_format.ParseDict(data, tasks_resp)

        logger.debug(
            f"Received tasks response - Round: {tasks_resp.round_number}, "
            f"Tasks: {[t.name for t in tasks_resp.tasks]}, "
            f"Sleep: {tasks_resp.sleep_time}, Quit: {tasks_resp.quit}"
        )
        return tasks_resp.tasks, tasks_resp.round_number, tasks_resp.sleep_time, tasks_resp.quit

    def get_aggregated_tensor(
        self,
        tensor_name: str,
        round_number: int,
        report: bool,
        tags: List[str],
        require_lossless: bool,
    ) -> Any:
        """Get aggregated tensor with proper security settings."""
        params = {
            "sender": self.collaborator_name,
            "receiver": self.aggregator_uuid,
            "federation_uuid": self.federation_uuid,
            "tensor_name": tensor_name,
            "round_number": round_number,
            "report": report,
            "tags": tags,
            "require_lossless": require_lossless,
            "collaborator_id": self.collaborator_name,
        }
        headers = {"Accept": "application/json", "Sender": self.collaborator_name}
        url = f"{self.base_url}/tensors/aggregated"
        extended_timeout = (30, 600)  # 30 seconds connect, 10 minutes read timeout
        try:
            logger.debug(f"Requesting aggregated tensor {tensor_name} for round {round_number}")
            response = self._make_request(
                "GET", url, params=params, headers=headers, timeout=extended_timeout
            )
            data = response.json()
            resp = aggregator_pb2.GetAggregatedTensorResponse()
            json_format.ParseDict(data, resp, ignore_unknown_fields=True)
            logger.debug(f"Successfully retrieved tensor {tensor_name} for round {round_number}")
            return resp.tensor
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                # This is expected during round 0 or when tensor hasn't been aggregated yet
                logger.debug(
                    f"No aggregated tensor found for {tensor_name} at round {round_number}"
                )
                return None
            raise

    def send_local_task_results(
        self,
        round_number: int,
        task_name: str,
        data_size: int,
        named_tensors: List[Any],
    ) -> bool:
        """Send local task results with proper security settings."""
        logger.debug(f"Sending task results for round {round_number}, task {task_name}")

        # Create the TaskResults message
        task_results = aggregator_pb2.TaskResults(
            header=aggregator_pb2.MessageHeader(
                sender=self.collaborator_name,
                receiver=self.aggregator_uuid,
                federation_uuid=self.federation_uuid,
                single_col_cert_common_name=self.single_col_cert_common_name or "",
            ),
            round_number=round_number,
            task_name=task_name,
            data_size=data_size,
            tensors=named_tensors,
        )

        # Serialize the TaskResults first
        task_results_bytes = task_results.SerializeToString()
        logger.debug(f"TaskResults serialized size: {len(task_results_bytes)} bytes")

        # Create a DataStream message containing the TaskResults bytes
        data_stream = base_pb2.DataStream(size=len(task_results_bytes), npbytes=task_results_bytes)

        # Create an empty DataStream to signal end of stream
        end_stream = base_pb2.DataStream(size=0, npbytes=b"")

        # Serialize both messages
        data_bytes = data_stream.SerializeToString()
        end_bytes = end_stream.SerializeToString()

        # Create length-prefixed stream format
        stream_data = (
            struct.pack(">I", len(data_bytes))  # Length prefix for first message
            + data_bytes  # First message
            + struct.pack(">I", len(end_bytes))  # Length prefix for second message
            + end_bytes  # Second message (empty message signals end)
        )

        url = f"{self.base_url}/tasks/results"
        request_headers = self._build_header()
        request_headers["Sender"] = self.collaborator_name
        request_headers["Content-Type"] = "application/x-protobuf-stream"
        request_headers["Content-Length"] = str(len(stream_data))

        try:
            response = self._make_request(
                "POST",
                url,
                data=stream_data,
                headers=request_headers,
                timeout=(30, 60),  # Keep shorter timeout since we're sending all data at once
            )
            response.raise_for_status()
            logger.debug(f"Successfully sent task results for round {round_number}")
            return True
        except Exception as e:
            logger.error(f"Failed to send task results for round {round_number}: {str(e)}")
            logger.error(f"Error type: {type(e).__name__}")
            logger.error(f"Request headers were: {request_headers}")
            raise

    def ping(self):
        """Ping the aggregator to check connectivity."""
        logger.info("Aggregator ping...")
        headers = {"Accept": "application/json", "Sender": self.collaborator_name}
        params = {
            "collaborator_id": self.collaborator_name,
            "federation_uuid": self.federation_uuid,
        }
        url = f"{self.base_url}/ping"
        response = self._make_request("GET", url, headers=headers, params=params)
        response.raise_for_status()
        data = response.json()

        # Validate response header like GRPC client
        header = data.get("header", {})
        assert header.get("receiver") == self.collaborator_name, (
            f"Receiver in response header does not match collaborator name. "
            f"Expected: {self.collaborator_name}, Actual: {header.get('receiver')}"
        )
        assert header.get("sender") == self.aggregator_uuid, (
            f"Sender in response header does not match aggregator UUID. "
            f"Expected: {self.aggregator_uuid}, Actual: {header.get('sender')}"
        )
        assert header.get("federationUuid") == self.federation_uuid, (
            f"Federation UUID in response header does not match. "
            f"Expected: {self.federation_uuid}, Actual: {header.get('federationUuid')}"
        )
        assert header.get("singleColCertCommonName", "") == (
            self.single_col_cert_common_name or ""
        ), (
            f"Single collaborator certificate common name in response header does not match. "
            f"Expected: {self.single_col_cert_common_name}, "
            f"Actual: {header.get('singleColCertCommonName')}"
        )

        logger.info("Aggregator pong!")

    def send_message_to_server(self, openfl_message: Any, collaborator_name: str) -> Any:
        """
        Forwards a converted message from the local REST client to the OpenFL server and returns
        the response.

        Args:
            openfl_message: The InteropMessage proto to be sent to the OpenFL server.
            collaborator_name: The name of the collaborator.

        Returns:
            The response from the OpenFL server (InteropMessage proto).
        """
        # Set the header fields
        header = aggregator_pb2.MessageHeader(
            sender=collaborator_name,
            receiver=self.aggregator_uuid,
            federation_uuid=self.federation_uuid,
            single_col_cert_common_name=self.single_col_cert_common_name or "",
        )
        openfl_message.header.CopyFrom(header)

        # Serialize to JSON
        json_payload = json_format.MessageToJson(openfl_message)
        url = f"{self.base_url}/interop/relay"
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Sender": collaborator_name,
        }
        response = self._make_request(
            "POST",
            url,
            data=json_payload,
            headers=headers,
            timeout=(30, 300),
        )
        response.raise_for_status()
        response_json = response.json()
        openfl_response = aggregator_pb2.InteropMessage()
        json_format.ParseDict(response_json, openfl_response, ignore_unknown_fields=True)
        return openfl_response

    def __del__(self):
        """Cleanup when the client is destroyed."""
        self.session.close()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.federated package."""

import os
from importlib import util

from openfl.federated.data import DataLoader  # NOQA
from openfl.federated.plan import Plan  # NOQA
from openfl.federated.task import TaskRunner  # NOQA

if util.find_spec("keras") is not None:
    from openfl.federated.data import KerasDataLoader
    from openfl.federated.task import KerasTaskRunner
if util.find_spec("torch") is not None:
    os.environ["SETUPTOOLS_USE_DISTUTILS"] = "stdlib"
    import sys

    import typing_extensions

    sys.modules["pip._vendor.typing_extensions"] = typing_extensions
    from openfl.federated.data import PyTorchDataLoader
    from openfl.federated.task import PyTorchTaskRunner
if util.find_spec("xgboost") is not None:
    from openfl.federated.data import XGBoostDataLoader
    from openfl.federated.task import XGBoostTaskRunner
if util.find_spec("flwr") is not None:
    from openfl.federated.data import FlowerDataLoader
    from openfl.federated.task import FlowerTaskRunner

__all__ = [
    "Plan",
    "TaskRunner",
    "DataLoader",
]



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from openfl.federated.data.sources.torch.folder_dataset import LabelMapper
from openfl.federated.data.sources.torch.image_folder import ImageFolder
from openfl.federated.data.sources.torch.verifiable_map_style_dataset import (
    VerifiableMapStyleDataset,
)


class VerifiableImageFolder(VerifiableMapStyleDataset):
    """VerifiableImageFolder class for image folder datasets."""

    def __init__(self, vds, transform=None, verify_dataset_items=False):
        self.label_mapper = LabelMapper()
        super().__init__(vds, transform=transform, verify_dataset_items=verify_dataset_items)

    def create_datasets(self):
        datasources = []
        for data_source in self.verifiable_dataset_info.data_sources:
            datasources.append(
                ImageFolder(data_source, self.label_mapper, transform=self.transform)
            )
        return datasources



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Framework Adapter plugin interface."""


class FrameworkAdapterPluginInterface:
    """Framework adapter plugin class."""

    def __init__(self) -> None:
        """Initialize framework adapter."""
        pass

    @staticmethod
    def serialization_setup():
        """Prepare model for serialization (optional)."""
        pass

    @staticmethod
    def get_tensor_dict(model, optimizer=None) -> dict:
        """Extract tensor dict from a model and an optimizer.

        Args:
            model (object): The model object.
            optimizer (object, optional): The optimizer object. Defaults to
                None.

        Returns:
            dict: A dictionary with weight name as key and numpy ndarray as
                value.

        Raises:
            NotImplementedError: This is a placeholder method that needs to be
                implemented in subclasses.
        """
        raise NotImplementedError

    @staticmethod
    def set_tensor_dict(model, tensor_dict, optimizer=None, device="cpu"):
        """
        Set tensor dict from a model and an optimizer.

        Given a dict {weight name: numpy ndarray} sets weights to
        the model and optimizer objects inplace.

        Args:
            model (object): The model object.
            tensor_dict (dict): Dictionary with weight name as key and numpy
                ndarray as value.
            optimizer (object, optional): The optimizer object. Defaults to
                None.
            device (str, optional): The device to be used. Defaults to 'cpu'.

        Returns:
            None

        Raises:
            NotImplementedError: This is a placeholder method that needs to be
                implemented in subclasses.
        """
        raise NotImplementedError



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Cryptography CA utilities."""

import datetime
import uuid
from typing import Tuple

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509.base import Certificate, CertificateSigningRequest
from cryptography.x509.extensions import ExtensionNotFound
from cryptography.x509.name import Name
from cryptography.x509.oid import ExtensionOID, NameOID


def generate_root_cert(
    days_to_expiration: int = 365,
) -> Tuple[RSAPrivateKey, Certificate]:
    """Generate a root certificate and its corresponding private key.

    Args:
        days_to_expiration (int, optional): The number of days until the
            certificate expires. Defaults to 365.

    Returns:
        Tuple[RSAPrivateKey, Certificate]: The private key and the certificate.
    """
    now = datetime.datetime.utcnow()
    expiration_delta = days_to_expiration * datetime.timedelta(1, 0, 0)

    # Generate private key
    root_private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=3072, backend=default_backend()
    )

    # Generate public key
    root_public_key = root_private_key.public_key()
    builder = x509.CertificateBuilder()
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.DOMAIN_COMPONENT, "org"),
            x509.NameAttribute(NameOID.DOMAIN_COMPONENT, "simple"),
            x509.NameAttribute(NameOID.COMMON_NAME, "Simple Root CA"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Simple Inc"),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "Simple Root CA"),
        ]
    )
    issuer = subject
    builder = builder.subject_name(subject)
    builder = builder.issuer_name(issuer)

    builder = builder.not_valid_before(now)
    builder = builder.not_valid_after(now + expiration_delta)
    builder = builder.serial_number(int(uuid.uuid4()))
    builder = builder.public_key(root_public_key)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=True, path_length=None),
        critical=True,
    )

    # Sign the CSR
    certificate = builder.sign(
        private_key=root_private_key,
        algorithm=hashes.SHA384(),
        backend=default_backend(),
    )

    return root_private_key, certificate


def generate_signing_csr() -> Tuple[RSAPrivateKey, CertificateSigningRequest]:
    """Generate a Certificate Signing Request (CSR) and its corresponding
    private key.

    Returns:
        Tuple[RSAPrivateKey, CertificateSigningRequest]: The private key and
            the CSR.
    """
    # Generate private key
    signing_private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=3072, backend=default_backend()
    )

    builder = x509.CertificateSigningRequestBuilder()
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.DOMAIN_COMPONENT, "org"),
            x509.NameAttribute(NameOID.DOMAIN_COMPONENT, "simple"),
            x509.NameAttribute(NameOID.COMMON_NAME, "Simple Signing CA"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Simple Inc"),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "Simple Signing CA"),
        ]
    )
    builder = builder.subject_name(subject)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=True, path_length=None),
        critical=True,
    )

    # Sign the CSR
    csr = builder.sign(
        private_key=signing_private_key,
        algorithm=hashes.SHA384(),
        backend=default_backend(),
    )

    return signing_private_key, csr


def sign_certificate(
    csr: CertificateSigningRequest,
    issuer_private_key: RSAPrivateKey,
    issuer_name: Name,
    days_to_expiration: int = 365,
    ca: bool = False,
) -> Certificate:
    """
    Sign a incoming Certificate Signing Request (CSR) with the issuer's
    private key.

    Args:
        csr (CertificateSigningRequest): The CSR to be signed.
        issuer_private_key (RSAPrivateKey): Root CA private key if the request
            is for the signing CA; Signing CA private key otherwise.
        issuer_name (Name): The name of the issuer.
        days_to_expiration (int, optional): The number of days until the
            certificate expires. Defaults to 365.
        ca (bool, optional): Whether the certificate is for a certificate
            authority (CA). Defaults to False.

    Returns:
        Certificate: The signed certificate.
    """
    now = datetime.datetime.utcnow()
    expiration_delta = days_to_expiration * datetime.timedelta(1, 0, 0)

    builder = x509.CertificateBuilder()
    builder = builder.subject_name(csr.subject)
    builder = builder.issuer_name(issuer_name)
    builder = builder.not_valid_before(now)
    builder = builder.not_valid_after(now + expiration_delta)
    builder = builder.serial_number(int(uuid.uuid4()))
    builder = builder.public_key(csr.public_key())
    builder = builder.add_extension(
        x509.BasicConstraints(ca=ca, path_length=None),
        critical=True,
    )
    try:
        builder = builder.add_extension(
            csr.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value,
            critical=False,
        )
    except ExtensionNotFound:
        pass  # Might not have alternative name

    signed_cert = builder.sign(
        private_key=issuer_private_key,
        algorithm=hashes.SHA384(),
        backend=default_backend(),
    )
    return signed_cert



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""This module contains the LocalDataSource class."""

import hashlib
import os
from hashlib import sha384
from pathlib import Path
from typing import Callable, Generator

from openfl.federated.data.sources.data_source import DataSource, DataSourceType


class LocalDataSource(DataSource):
    """This class represents a local data source."""

    def __init__(
        self,
        name: str,
        source_path: Path,
        base_path,
        hash_func: Callable[..., "hashlib._Hash"] = sha384,
        max_dataset_size=0,
    ):
        """
        Initialize a LocalDataSource object.

        Args:
            name (str): The name of the data source.
            source_path (Path): The path to the source data, relative to base_path.
            base_path (Path): The base path to the data source.
            hash_func (Callable[..., hashlib._Hash]): The hash function from hashlib
            to use to hash the data.
            max_dataset_size (int): The maximum size of the dataset in GB.
        """
        super().__init__(DataSourceType.LOCAL, name)
        self.source_path = Path(source_path)
        if not super().is_valid_hash_function(hash_func):
            raise ValueError(
                f"Data source {self.name}: Invalid hash function: {hash_func.__name__}."
                " Must be a hashlib function."
            )
        self.hash_func = hash_func
        self.max_dataset_size = max_dataset_size
        self._base_path = Path(base_path)  # private attribute, will not be serialized

    def get_source_full_path(self):
        """Return the full path to the source data."""
        return self._base_path / self.source_path

    def enumerate_files(self) -> Generator[str, None, None]:
        """Enumerate all files in the data source."""
        total_size_bytes = 0
        full_path = Path(self._base_path) / self.source_path
        if full_path.is_dir():
            for root, _, files in os.walk(full_path):
                for file in files:
                    file_path = Path(root) / file
                    if self.max_dataset_size > 0:
                        total_size_bytes += file_path.stat().st_size
                        total_size_gb = total_size_bytes / (1024**3)
                        if total_size_gb > self.max_dataset_size:
                            raise ValueError(
                                f"Data source {self.name}: Total dataset size: {total_size_gb:.2f}"
                                f" GB exceeds {self.max_dataset_size} GB"
                            )
                    yield str(file_path)

        elif full_path.is_file():
            if self.max_dataset_size > 0:
                total_size_bytes = full_path.stat().st_size
                total_size_gb = total_size_bytes / (1024**3)
                if total_size_gb > self.max_dataset_size:
                    raise ValueError(
                        f"Data source {self.name}: Total dataset size: {total_size_gb:.2f}"
                        f" GB exceeds {self.max_dataset_size} GB"
                    )
            yield str(full_path)

    def compute_file_hash(self, path: str) -> str:
        """Compute the hash of the file. Return hash on hexstring format."""
        hash_obj = self.hash_func()
        with open(path, "rb") as file:
            for byte_block in iter(lambda: file.read(65536), b""):
                hash_obj.update(byte_block)
        return hash_obj.hexdigest()

    def read_blob(self, path: str) -> bytes:
        """Read a blob from the data source."""
        with open(path, "rb") as file:
            return file.read()

    @classmethod
    def from_dict(cls, ds_dict: dict, base_path):
        source_path = Path(ds_dict["source_path"])
        # Retrieve function from hashlib
        hash_func = getattr(hashlib, ds_dict.get("hash_func", "sha384"), None)
        max_dataset_size = ds_dict.get("max_dataset_size", 0)
        return cls(
            name=ds_dict["name"],
            source_path=source_path,
            base_path=base_path,
            hash_func=hash_func,
            max_dataset_size=max_dataset_size,
        )



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.utilities package."""

from openfl.experimental.workflow.utilities.exceptions import (
    ResourcesAllocationError,
    ResourcesNotAvailableError,
    SerializationError,
)
from openfl.experimental.workflow.utilities.metaflow_utils import MetaflowInterface
from openfl.experimental.workflow.utilities.resources import get_number_of_gpus
from openfl.experimental.workflow.utilities.runtime_utils import (
    check_resource_allocation,
    checkpoint,
    filter_attributes,
    generate_artifacts,
    parse_attrs,
)
from openfl.experimental.workflow.utilities.stream_redirect import (
    RedirectStdStream,
    RedirectStdStreamBuffer,
    RedirectStdStreamContext,
)
from openfl.experimental.workflow.utilities.transitions import (
    aggregator_to_collaborator,
    collaborator_to_aggregator,
    should_transfer,
)



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from flwr.proto import grpcadapter_pb2

from openfl.protocols import aggregator_pb2


def flower_to_openfl_message(flower_message, header=None, end_experiment=False):
    """
    Convert a Flower MessageContainer to an OpenFL InteropMessage.

    This function takes a Flower MessageContainer and converts it into an OpenFL InteropMessage.
    If the input is already an OpenFL InteropMessage, it returns the input as-is.

    Args:
        flower_message (grpcadapter_pb2.MessageContainer or aggregator_pb2.InteropMessage):
            The Flower message to be converted. It can either be a Flower MessageContainer
            or an OpenFL InteropMessage.
        header (aggregator_pb2.MessageHeader, optional):
            An optional header to be included in the OpenFL InteropMessage. If provided,
            it will be copied to the InteropMessage's header field.

    Returns:
        aggregator_pb2.InteropMessage: The converted OpenFL InteropMessage message.
    """
    if isinstance(flower_message, aggregator_pb2.InteropMessage):
        # If the input is already an OpenFL message, return it as-is
        return flower_message
    else:
        # Create the OpenFL message
        openfl_message = aggregator_pb2.InteropMessage()
        # Set the MessageHeader fields based on the provided sender and receiver
        if header:
            openfl_message.header.CopyFrom(header)

        # Serialize the Flower message and set it in the OpenFL message
        serialized_flower_message = flower_message.SerializeToString()
        openfl_message.message.npbytes = serialized_flower_message
        openfl_message.message.size = len(serialized_flower_message)

        # Add flag to check if experiment has ended
        openfl_message.metadata.update({"end_experiment": str(end_experiment)})
        return openfl_message


def openfl_to_flower_message(openfl_message):
    """
    Convert an OpenFL InteropMessage to a Flower MessageContainer.

    This function takes an OpenFL InteropMessage and converts it into a Flower MessageContainer.
    If the input is already a Flower MessageContainer, it returns the input as-is.

    Args:
        openfl_message (aggregator_pb2.InteropMessage or grpcadapter_pb2.MessageContainer):
            The OpenFL message to be converted. It can either be an OpenFL InteropMessage
            or a Flower MessageContainer.

    Returns:
        grpcadapter_pb2.MessageContainer: The converted Flower MessageContainer.
    """
    if isinstance(openfl_message, grpcadapter_pb2.MessageContainer):
        # If the input is already a Flower message, return it as-is
        return openfl_message
    else:
        # Deserialize the Flower message from the DataStream npbytes field
        flower_message = grpcadapter_pb2.MessageContainer()
        flower_message.ParseFromString(openfl_message.message.npbytes)
        return flower_message



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Notebook Tools module."""

import shutil
from importlib import import_module
from pathlib import Path
from shutil import copytree
from typing import Any, Dict, Tuple

from openfl.experimental.workflow.federated.plan import Plan
from openfl.experimental.workflow.interface.cli.cli_helper import WORKSPACE, print_tree
from openfl.experimental.workflow.notebooktools.code_analyzer import CodeAnalyzer


class NotebookTools:
    """Class providing utility functions to convert Jupyter notebook based on Workflow API
    into a workspace enabling its deployment on distributed infrastructure

    Attributes:
        notebook_path (Path): Path to Jupyter notebook to be converted.
        output_workspace_path (Path): Target directory for generated workspace.
        template_workspace_path (Path): Path to template workspace provided with
            OpenFL.
        code_analyzer (CodeAnalyzer): An instance of the CodeAnalyzer class for analyzing
            notebook code.
        flow_class_name (str): Name of the flow class.
    """

    def __init__(self, notebook_path: str, output_workspace: str) -> None:
        """Initialize a NotebookTools object.
        Args:
            notebook_path (str): Path to Jupyter notebook to be converted.
            output_workspace (str): Target directory for generated workspace
        """
        self.notebook_path = Path(notebook_path).resolve()
        if not self.notebook_path.exists() or not self.notebook_path.is_file():
            raise FileNotFoundError(f"The Jupyter notebook at {notebook_path} does not exist.")

        self.output_workspace_path = Path(output_workspace).resolve()
        self._initialize_workspace()

        self.code_analyzer = CodeAnalyzer(self.notebook_path, self.output_workspace_path)

    def _initialize_workspace(self) -> None:
        """Initialize the workspace by setting up path and copying templates"""
        # Regenerate the workspace if it already exists
        if self.output_workspace_path.exists():
            print(f"Removing existing workspace: {self.output_workspace_path}")
            shutil.rmtree(self.output_workspace_path)
        self.output_workspace_path.parent.mkdir(parents=True, exist_ok=True)

        self.template_workspace_path = (WORKSPACE / "template_workspace").resolve(strict=True)
        copytree(self.template_workspace_path, self.output_workspace_path)
        print(f"Copied template workspace to {self.output_workspace_path}")

    def _generate_experiment_archive(self) -> Tuple[str, str]:
        """
        Create archive of the generated workspace

        Returns:
            Tuple[str, str]: A tuple containing:
                (archive_path, flow_class_name).
        """
        parent_directory = self.output_workspace_path.parent
        archive_path = parent_directory / "experiment"
        # Create a ZIP archive of the generated_workspace directory
        arch_path = shutil.make_archive(str(archive_path), "zip", str(self.output_workspace_path))
        print(f"Archive created at {archive_path}.zip")

        return arch_path, self.flow_class_name

    def _generate_requirements(self) -> None:
        """Extract dependencies (pip install <module name>) from exported python script
        and append to workspace/requirements.txt
        """
        try:
            requirements_filepath = str(
                self.output_workspace_path.joinpath("requirements.txt").resolve()
            )
            with open(requirements_filepath, "a") as f:
                f.writelines(self.code_analyzer.requirements)

            print(f"Successfully generated {requirements_filepath}")

        except Exception as e:
            print(f"Failed to generate requirements: {e}")

    def _generate_plan_yaml(self, director_fqdn: str = None, tls: bool = False) -> None:
        """Generates workspace/plan.yaml
        Args:
            director_fqdn (str): Fully qualified domain name of the director node.
            tls (bool, optional): Whether to use TLS for the connection.
        """
        flow_details = self._extract_flow_details()
        flow_config = self.code_analyzer.fetch_flow_configuration(flow_details)
        plan_path = self.output_workspace_path.joinpath("plan", "plan.yaml").resolve()
        data_config = self._build_plan_config(flow_config, director_fqdn, tls, plan_path)
        Plan.dump(plan_path, data_config)

    def _build_plan_config(
        self, flow_config: Dict[str, Any], director_fqdn: str, tls: bool, plan_path: Path
    ) -> Dict[str, Any]:
        """
        Build plan configuration with validation.

        Args:
            flow_config: Flow configuration dictionary
            director_fqdn: Director's FQDN
            tls: TLS setting
            plan_path: Path to plan.yaml

        Returns:
            Dict[str, Any]: Complete plan configuration
        """
        data_config = self._initialize_plan_yaml(plan_path)
        data_config["federated_flow"].update(flow_config["federated_flow"])
        data_config["network"]["settings"]["agg_addr"] = director_fqdn
        data_config["network"]["settings"]["tls"] = tls

        return data_config

    def _extract_flow_details(self) -> str:
        """Extract the flow class details"""
        flspsec = import_module("openfl.experimental.workflow.interface").FLSpec
        flow_details = self.code_analyzer.get_flow_class_details(flspsec)
        # Set flow_class_name for future reference
        self.flow_class_name = flow_details["flow_class_name"]

        return flow_details

    def _initialize_plan_yaml(self, plan_yaml: Path) -> dict:
        """Load or initialize the plan YAML data.
        Args:
            plan_yaml (Path): The path to the plan.yaml file.

        Returns:
            dict: The data dictionary from plan.yaml.
        """
        data = Plan.load(plan_yaml)
        if data is None:
            data = {}
            data["federated_flow"] = {"settings": {}, "template": ""}

        return data

    def export(self, director_fqdn: str, tls: bool = False) -> Tuple[str, str]:
        """Exports workspace for FederatedRuntime.
        Args:
            director_fqdn (str): Fully qualified domain name of the director node.
            tls (bool, optional): Whether to use TLS for the connection.

        Returns:
            Tuple[str, str]: A tuple containing:
                (archive_path, flow_class_name).
        """
        self._generate_requirements()
        self._generate_plan_yaml(director_fqdn, tls)
        print_tree(self.output_workspace_path, level=2)

        return self._generate_experiment_archive()
